# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import itertools
import os
import os.path as osp
import pathlib
import pickle
import signal
import subprocess
import sys
import threading
import time
import typing as t

from smartredis import Client, ConfigOptions

from smartsim._core.utils.network import get_ip_from_host

from ..._core.launcher.step import Step
from ..._core.utils.helpers import (
    SignalInterceptionStack,
    unpack_colo_db_identifier,
    unpack_db_identifier,
)
from ..._core.utils.redis import (
    db_is_active,
    set_ml_model,
    set_script,
    shutdown_db_node,
)
from ...database import Orchestrator
from ...entity import Ensemble, EntitySequence, Model, SmartSimEntity
from ...error import (
    LauncherError,
    SmartSimError,
    SSDBIDConflictError,
    SSInternalError,
    SSUnsupportedError,
)
from ...log import get_logger
from ...servertype import CLUSTERED, STANDALONE
from ...status import TERMINAL_STATUSES, SmartSimStatus
from ..config import CONFIG
from ..launcher import (
    DragonLauncher,
    LocalLauncher,
    LSFLauncher,
    PBSLauncher,
    SlurmLauncher,
)
from ..launcher.launcher import Launcher
from ..utils import check_cluster_status, create_cluster, serialize
from .controller_utils import _AnonymousBatchJob, _look_up_launched_data
from .job import Job
from .jobmanager import JobManager
from .manifest import LaunchedManifest, LaunchedManifestBuilder, Manifest

if t.TYPE_CHECKING:
    from types import FrameType

    from ..utils.serialize import TStepLaunchMetaData


logger = get_logger(__name__)

# job manager lock
JM_LOCK = threading.RLock()


class Controller:
    """The controller module provides an interface between the
    smartsim entities created in the experiment and the
    underlying workload manager or run framework.
    """

    def __init__(self, launcher: str = "local") -> None:
        """Initialize a Controller

        :param launcher: the type of launcher being used
        """
        self._jobs = JobManager(JM_LOCK)
        self.init_launcher(launcher)
        self._telemetry_monitor: t.Optional[subprocess.Popen[bytes]] = None

    def start(
        self,
        exp_name: str,
        exp_path: str,
        manifest: Manifest,
        block: bool = True,
        kill_on_interrupt: bool = True,
    ) -> None:
        """Start the passed SmartSim entities

        This function should not be called directly, but rather
        through the experiment interface.

        The controller will start the job-manager thread upon
        execution of all jobs.
        """
        # launch a telemetry monitor to track job progress
        if CONFIG.telemetry_enabled:
            self._start_telemetry_monitor(exp_path)

        self._jobs.kill_on_interrupt = kill_on_interrupt

        # register custom signal handler for ^C (SIGINT)
        SignalInterceptionStack.get(signal.SIGINT).push_unique(
            self._jobs.signal_interrupt
        )
        launched = self._launch(exp_name, exp_path, manifest)

        # start the job manager thread if not already started
        if not self._jobs.actively_monitoring:
            self._jobs.start()

        serialize.save_launch_manifest(
            launched.map(_look_up_launched_data(self._launcher))
        )

        # block until all non-database jobs are complete
        if block:
            # poll handles its own keyboard interrupt as
            # it may be called separately
            self.poll(5, True, kill_on_interrupt=kill_on_interrupt)

    @property
    def active_orchestrator_jobs(self) -> t.Dict[str, Job]:
        """Return active orchestrator jobs."""
        return {**self._jobs.db_jobs}

    @property
    def orchestrator_active(self) -> bool:
        with JM_LOCK:
            if len(self._jobs.db_jobs) > 0:
                return True
            return False

    def poll(
        self, interval: int, verbose: bool, kill_on_interrupt: bool = True
    ) -> None:
        """Poll running jobs and receive logging output of job status

        :param interval: number of seconds to wait before polling again
        :param verbose: set verbosity
        :param kill_on_interrupt: flag for killing jobs when SIGINT is received
        """
        self._jobs.kill_on_interrupt = kill_on_interrupt
        to_monitor = self._jobs.jobs
        while len(to_monitor) > 0:
            time.sleep(interval)

            # acquire lock to avoid "dictionary changed during iteration" error
            # without having to copy dictionary each time.
            if verbose:
                with JM_LOCK:
                    for job in to_monitor.values():
                        logger.info(job)

    def finished(
        self, entity: t.Union[SmartSimEntity, EntitySequence[SmartSimEntity]]
    ) -> bool:
        """Return a boolean indicating wether a job has finished or not

        :param entity: object launched by SmartSim.
        :returns: bool
        :raises ValueError: if entity has not been launched yet
        """
        try:
            if isinstance(entity, Orchestrator):
                raise TypeError("Finished() does not support Orchestrator instances")
            if isinstance(entity, EntitySequence):
                return all(self.finished(ent) for ent in entity.entities)
            if not isinstance(entity, SmartSimEntity):
                raise TypeError(
                    f"Argument was of type {type(entity)} not derived "
                    "from SmartSimEntity or EntitySequence"
                )

            return self._jobs.is_finished(entity)
        except KeyError:
            raise ValueError(
                f"Entity {entity.name} has not been launched in this experiment"
            ) from None

    def stop_entity(
        self, entity: t.Union[SmartSimEntity, EntitySequence[SmartSimEntity]]
    ) -> None:
        """Stop an instance of an entity

        This function will also update the status of the job in
        the jobmanager so that the job appears as "cancelled".

        :param entity: entity to be stopped
        """
        with JM_LOCK:
            job = self._jobs[entity.name]
            if job.status not in TERMINAL_STATUSES:
                logger.info(
                    " ".join(
                        ("Stopping model", entity.name, "with job name", str(job.name))
                    )
                )
                status = self._launcher.stop(job.name)

                job.set_status(
                    status.status,
                    status.launcher_status,
                    status.returncode,
                    error=status.error,
                    output=status.output,
                )
                self._jobs.move_to_completed(job)

    def stop_db(self, db: Orchestrator) -> None:
        """Stop an orchestrator

        :param db: orchestrator to be stopped
        """
        if db.batch:
            self.stop_entity(db)
        else:
            with JM_LOCK:
                for node in db.entities:
                    for host_ip, port in itertools.product(
                        (get_ip_from_host(host) for host in node.hosts), db.ports
                    ):
                        retcode, _, _ = shutdown_db_node(host_ip, port)
                        # Sometimes the DB will not shutdown (unless we force NOSAVE)
                        if retcode != 0:
                            self.stop_entity(node)
                            continue

                        job = self._jobs[node.name]
                        job.set_status(
                            SmartSimStatus.STATUS_CANCELLED,
                            "",
                            0,
                            output=None,
                            error=None,
                        )
                        self._jobs.move_to_completed(job)

        db.reset_hosts()

    def stop_entity_list(self, entity_list: EntitySequence[SmartSimEntity]) -> None:
        """Stop an instance of an entity list

        :param entity_list: entity list to be stopped
        """

        if entity_list.batch:
            self.stop_entity(entity_list)
        else:
            for entity in entity_list.entities:
                self.stop_entity(entity)

    def get_jobs(self) -> t.Dict[str, Job]:
        """Return a dictionary of completed job data

        :returns: dict[str, Job]
        """
        with JM_LOCK:
            return self._jobs.completed

    def get_entity_status(
        self, entity: t.Union[SmartSimEntity, EntitySequence[SmartSimEntity]]
    ) -> SmartSimStatus:
        """Get the status of an entity

        :param entity: entity to get status of
        :raises TypeError: if not SmartSimEntity | EntitySequence
        :return: status of entity
        """
        if not isinstance(entity, (SmartSimEntity, EntitySequence)):
            raise TypeError(
                "Argument must be of type SmartSimEntity or EntitySequence, "
                f"not {type(entity)}"
            )
        return self._jobs.get_status(entity)

    def get_entity_list_status(
        self, entity_list: EntitySequence[SmartSimEntity]
    ) -> t.List[SmartSimStatus]:
        """Get the statuses of an entity list

        :param entity_list: entity list containing entities to
                            get statuses of
        :raises TypeError: if not EntitySequence
        :return: list of SmartSimStatus statuses
        """
        if not isinstance(entity_list, EntitySequence):
            raise TypeError(
                f"Argument was of type {type(entity_list)} not EntitySequence"
            )
        if entity_list.batch:
            return [self.get_entity_status(entity_list)]
        statuses = []
        for entity in entity_list.entities:
            statuses.append(self.get_entity_status(entity))
        return statuses

    def init_launcher(self, launcher: str) -> None:
        """Initialize the controller with a specific type of launcher.
        SmartSim currently supports slurm, pbs(pro), lsf,
        and local launching

        :param launcher: which launcher to initialize
        :raises SSUnsupportedError: if a string is passed that is not
                                    a supported launcher
        :raises TypeError: if no launcher argument is provided.
        """
        launcher_map: t.Dict[str, t.Type[Launcher]] = {
            "slurm": SlurmLauncher,
            "pbs": PBSLauncher,
            "pals": PBSLauncher,
            "lsf": LSFLauncher,
            "local": LocalLauncher,
            "dragon": DragonLauncher,
        }

        if launcher is not None:
            launcher = launcher.lower()
            if launcher in launcher_map:
                # create new instance of the launcher
                self._launcher = launcher_map[launcher]()
                self._jobs.set_launcher(self._launcher)
            else:
                raise SSUnsupportedError("Launcher type not supported: " + launcher)
        else:
            raise TypeError("Must provide a 'launcher' argument")

    @staticmethod
    def symlink_output_files(
        job_step: Step, entity: t.Union[SmartSimEntity, EntitySequence[SmartSimEntity]]
    ) -> None:
        """Create symlinks for entity output files that point to the output files
        under the .smartsim directory

        :param job_step: Job step instance
        :param entity: Entity instance
        """
        historical_out, historical_err = map(pathlib.Path, job_step.get_output_files())
        entity_out = pathlib.Path(entity.path) / f"{entity.name}.out"
        entity_err = pathlib.Path(entity.path) / f"{entity.name}.err"

        # check if there is already a link to a previous run
        if entity_out.is_symlink() or entity_err.is_symlink():
            entity_out.unlink()
            entity_err.unlink()

        historical_err.touch()
        historical_out.touch()

        if historical_err.exists() and historical_out.exists():
            entity_out.symlink_to(historical_out)
            entity_err.symlink_to(historical_err)
        else:
            raise FileNotFoundError(
                f"Output files for {entity.name} could not be found. "
                "Symlinking files failed."
            )

    def _launch(
        self, exp_name: str, exp_path: str, manifest: Manifest
    ) -> LaunchedManifest[t.Tuple[str, Step]]:
        """Main launching function of the controller

        Orchestrators are always launched first so that the
        address of the database can be given to following entities

        :param exp_name: The name of the launching experiment
        :param exp_path: path to location of ``Experiment`` directory if generated
        :param manifest: Manifest of deployables to launch
        """

        manifest_builder = LaunchedManifestBuilder[t.Tuple[str, Step]](
            exp_name=exp_name,
            exp_path=exp_path,
            launcher_name=str(self._launcher),
        )
        # Loop over deployables to launch and launch multiple orchestrators
        for orchestrator in manifest.dbs:
            for key in self._jobs.get_db_host_addresses():
                _, db_id = unpack_db_identifier(key, "_")
                if orchestrator.db_identifier == db_id:
                    raise SSDBIDConflictError(
                        f"Database identifier {orchestrator.db_identifier}"
                        " has already been used. Pass in a unique"
                        " name for db_identifier"
                    )

            if orchestrator.num_shards > 1 and isinstance(
                self._launcher, LocalLauncher
            ):
                raise SmartSimError(
                    "Local launcher does not support multi-host orchestrators"
                )
            self._launch_orchestrator(orchestrator, manifest_builder)

        if self.orchestrator_active:
            self._set_dbobjects(manifest)

        # create all steps prior to launch
        steps: t.List[
            t.Tuple[Step, t.Union[SmartSimEntity, EntitySequence[SmartSimEntity]]]
        ] = []

        symlink_substeps: t.List[
            t.Tuple[Step, t.Union[SmartSimEntity, EntitySequence[SmartSimEntity]]]
        ] = []

        for elist in manifest.ensembles:
            ens_telem_dir = manifest_builder.run_telemetry_subdirectory / "ensemble"
            if elist.batch:
                batch_step, substeps = self._create_batch_job_step(elist, ens_telem_dir)
                manifest_builder.add_ensemble(
                    elist, [(batch_step.name, step) for step in substeps]
                )

                # symlink substeps to maintain directory structure
                for substep, substep_entity in zip(substeps, elist.models):
                    symlink_substeps.append((substep, substep_entity))

                steps.append((batch_step, elist))
            else:
                # if ensemble is to be run as separate job steps, aka not in a batch
                job_steps = [
                    (self._create_job_step(e, ens_telem_dir / elist.name), e)
                    for e in elist.entities
                ]
                manifest_builder.add_ensemble(
                    elist, [(step.name, step) for step, _ in job_steps]
                )
                steps.extend(job_steps)
        # models themselves cannot be batch steps. If batch settings are
        # attached, wrap them in an anonymous batch job step
        for model in manifest.models:
            model_telem_dir = manifest_builder.run_telemetry_subdirectory / "model"
            if model.batch_settings:
                anon_entity_list = _AnonymousBatchJob(model)
                batch_step, substeps = self._create_batch_job_step(
                    anon_entity_list, model_telem_dir
                )
                manifest_builder.add_model(model, (batch_step.name, batch_step))

                symlink_substeps.append((substeps[0], model))
                steps.append((batch_step, model))
            else:
                job_step = self._create_job_step(model, model_telem_dir)
                manifest_builder.add_model(model, (job_step.name, job_step))
                steps.append((job_step, model))

        # launch and symlink steps
        for step, entity in steps:
            self._launch_step(step, entity)
            self.symlink_output_files(step, entity)

        # symlink substeps to maintain directory structure
        for substep, entity in symlink_substeps:
            self.symlink_output_files(substep, entity)

        return manifest_builder.finalize()

    def _launch_orchestrator(
        self,
        orchestrator: Orchestrator,
        manifest_builder: LaunchedManifestBuilder[t.Tuple[str, Step]],
    ) -> None:
        """Launch an Orchestrator instance

        This function will launch the Orchestrator instance and
        if on WLM, find the nodes where it was launched and
        set them in the JobManager

        :param orchestrator: orchestrator to launch
        :param manifest_builder: An `LaunchedManifestBuilder` to record the
                                 names and `Step`s of the launched orchestrator
        """
        orchestrator.remove_stale_files()
        orc_telem_dir = manifest_builder.run_telemetry_subdirectory / "database"

        # if the orchestrator was launched as a batch workload
        if orchestrator.batch:
            orc_batch_step, substeps = self._create_batch_job_step(
                orchestrator, orc_telem_dir
            )
            manifest_builder.add_database(
                orchestrator, [(orc_batch_step.name, step) for step in substeps]
            )

            self._launch_step(orc_batch_step, orchestrator)
            self.symlink_output_files(orc_batch_step, orchestrator)

            # symlink substeps to maintain directory structure
            for substep, substep_entity in zip(substeps, orchestrator.entities):
                self.symlink_output_files(substep, substep_entity)

        # if orchestrator was run on existing allocation, locally, or in allocation
        else:
            db_steps = [
                (self._create_job_step(db, orc_telem_dir / orchestrator.name), db)
                for db in orchestrator.entities
            ]
            manifest_builder.add_database(
                orchestrator, [(step.name, step) for step, _ in db_steps]
            )
            for db_step in db_steps:
                self._launch_step(*db_step)
                self.symlink_output_files(*db_step)

        # wait for orchestrator to spin up
        self._orchestrator_launch_wait(orchestrator)

        # set the jobs in the job manager to provide SSDB variable to entities
        # if _host isnt set within each
        self._jobs.set_db_hosts(orchestrator)

        # create the database cluster
        if orchestrator.num_shards > 2:
            num_trials = 5
            cluster_created = False
            while not cluster_created:
                try:
                    create_cluster(orchestrator.hosts, orchestrator.ports)
                    check_cluster_status(orchestrator.hosts, orchestrator.ports)
                    num_shards = orchestrator.num_shards
                    logger.info(f"Database cluster created with {num_shards} shards")
                    cluster_created = True
                except SSInternalError:
                    if num_trials > 0:
                        logger.debug(
                            "Cluster creation failed, attempting again in five seconds."
                        )
                        num_trials -= 1
                        time.sleep(5)
                    else:
                        # surface SSInternalError as we have no way to recover
                        raise
        self._save_orchestrator(orchestrator)
        logger.debug(f"Orchestrator launched on nodes: {orchestrator.hosts}")

    def _launch_step(
        self,
        job_step: Step,
        entity: t.Union[SmartSimEntity, EntitySequence[SmartSimEntity]],
    ) -> None:
        """Use the launcher to launch a job step

        :param job_step: a job step instance
        :param entity: entity instance
        :raises SmartSimError: if launch fails
        """
        # attempt to retrieve entity name in JobManager.completed
        completed_job = self._jobs.completed.get(entity.name, None)

        # if completed job DNE and is the entity name is not
        # running in JobManager.jobs or JobManager.db_jobs,
        # launch the job
        if completed_job is None and (
            entity.name not in self._jobs.jobs and entity.name not in self._jobs.db_jobs
        ):
            try:
                job_id = self._launcher.run(job_step)
            except LauncherError as e:
                msg = f"An error occurred when launching {entity.name} \n"
                msg += "Check error and output files for details.\n"
                msg += f"{entity}"
                logger.error(msg)
                raise SmartSimError(f"Job step {entity.name} failed to launch") from e

        # if the completed job does exist and the entity passed in is the same
        # that has ran and completed, relaunch the entity.
        elif completed_job is not None and completed_job.entity is entity:
            try:
                job_id = self._launcher.run(job_step)
            except LauncherError as e:
                msg = f"An error occurred when launching {entity.name} \n"
                msg += "Check error and output files for details.\n"
                msg += f"{entity}"
                logger.error(msg)
                raise SmartSimError(f"Job step {entity.name} failed to launch") from e

        # the entity is using a duplicate name of an existing entity in
        # the experiment, throw an error
        else:
            raise SSUnsupportedError("SmartSim entities cannot have duplicate names.")

        # a job step is a task if it is not managed by a workload manager (i.e. Slurm)
        # but is rather started, monitored, and exited through the Popen interface
        # in the taskmanager
        is_task = not job_step.managed

        if self._jobs.query_restart(entity.name):
            logger.debug(f"Restarting {entity.name}")
            self._jobs.restart_job(job_step.name, job_id, entity.name, is_task)
        else:
            logger.debug(f"Launching {entity.name}")
            self._jobs.add_job(job_step.name, job_id, entity, is_task)

    def _create_batch_job_step(
        self,
        entity_list: t.Union[Orchestrator, Ensemble, _AnonymousBatchJob],
        telemetry_dir: pathlib.Path,
    ) -> t.Tuple[Step, t.List[Step]]:
        """Use launcher to create batch job step

        :param entity_list: EntityList to launch as batch
        :param telemetry_dir: Path to a directory in which the batch job step
                              may write telemetry events
        :return: batch job step instance and a list of run steps to be
                 executed within the batch job
        """
        if not entity_list.batch_settings:
            raise ValueError(
                "EntityList must have batch settings to be launched as batch"
            )

        telemetry_dir = telemetry_dir / entity_list.name
        batch_step = self._launcher.create_step(
            entity_list.name, entity_list.path, entity_list.batch_settings
        )
        batch_step.meta["entity_type"] = str(type(entity_list).__name__).lower()
        batch_step.meta["status_dir"] = str(telemetry_dir)

        substeps = []
        for entity in entity_list.entities:
            # tells step creation not to look for an allocation
            entity.run_settings.in_batch = True
            step = self._create_job_step(entity, telemetry_dir)
            substeps.append(step)
            batch_step.add_to_batch(step)
        return batch_step, substeps

    def _create_job_step(
        self, entity: SmartSimEntity, telemetry_dir: pathlib.Path
    ) -> Step:
        """Create job steps for all entities with the launcher

        :param entity: an entity to create a step for
        :param telemetry_dir: Path to a directory in which the job step
                               may write telemetry events
        :return: the job step
        """
        # get SSDB, SSIN, SSOUT and add to entity run settings
        if isinstance(entity, Model):
            self._prep_entity_client_env(entity)

        step = self._launcher.create_step(entity.name, entity.path, entity.run_settings)

        step.meta["entity_type"] = str(type(entity).__name__).lower()
        step.meta["status_dir"] = str(telemetry_dir / entity.name)

        return step

    def _prep_entity_client_env(self, entity: Model) -> None:
        """Retrieve all connections registered to this entity

        :param entity: The entity to retrieve connections from
        """

        client_env: t.Dict[str, t.Union[str, int, float, bool]] = {}
        address_dict = self._jobs.get_db_host_addresses()

        for db_id, addresses in address_dict.items():
            db_name, _ = unpack_db_identifier(db_id, "_")
            if addresses:
                # Cap max length of SSDB
                client_env[f"SSDB{db_name}"] = ",".join(addresses[:128])

                # Retrieve num_shards to append to client env
                client_env[f"SR_DB_TYPE{db_name}"] = (
                    CLUSTERED if len(addresses) > 1 else STANDALONE
                )

        if entity.incoming_entities:
            client_env["SSKEYIN"] = ",".join(
                [in_entity.name for in_entity in entity.incoming_entities]
            )
        if entity.query_key_prefixing():
            client_env["SSKEYOUT"] = entity.name

        # Set address to local if it's a colocated model
        if entity.colocated and entity.run_settings.colocated_db_settings is not None:
            db_name_colo = entity.run_settings.colocated_db_settings["db_identifier"]
            assert isinstance(db_name_colo, str)
            for key in address_dict:
                _, db_id = unpack_db_identifier(key, "_")
                if db_name_colo == db_id:
                    raise SSDBIDConflictError(
                        f"Database identifier {db_name_colo}"
                        " has already been used. Pass in a unique"
                        " name for db_identifier"
                    )

            db_name_colo = unpack_colo_db_identifier(db_name_colo)
            if colo_cfg := entity.run_settings.colocated_db_settings:
                port = colo_cfg.get("port", None)
                socket = colo_cfg.get("unix_socket", None)
                if socket and port:
                    raise SSInternalError(
                        "Co-located was configured for both TCP/IP and UDS"
                    )
                if port:
                    client_env[f"SSDB{db_name_colo}"] = f"127.0.0.1:{str(port)}"
                elif socket:
                    client_env[f"SSDB{db_name_colo}"] = f"unix://{socket}"
                else:
                    raise SSInternalError(
                        "Colocated database was not configured for either TCP or UDS"
                    )
                client_env[f"SR_DB_TYPE{db_name_colo}"] = STANDALONE

        entity.run_settings.update_env(client_env)

    def _save_orchestrator(self, orchestrator: Orchestrator) -> None:
        """Save the orchestrator object via pickle

        This function saves the orchestrator information to a pickle
        file that can be imported by subsequent experiments to reconnect
        to the orchestrator.

        :param orchestrator: Orchestrator configuration to be saved
        """

        if not orchestrator.is_active():
            raise Exception("Orchestrator is not running")

        # Extract only the db_jobs associated with this particular orchestrator
        if orchestrator.batch:
            job_names = [orchestrator.name]
        else:
            job_names = [dbnode.name for dbnode in orchestrator.entities]
        db_jobs = {
            name: job for name, job in self._jobs.db_jobs.items() if name in job_names
        }

        # Extract the associated steps
        steps = [
            self._launcher.step_mapping[db_job.name] for db_job in db_jobs.values()
        ]

        orc_data = {"db": orchestrator, "db_jobs": db_jobs, "steps": steps}

        with open(orchestrator.checkpoint_file, "wb") as pickle_file:
            pickle.dump(orc_data, pickle_file)

    def _orchestrator_launch_wait(self, orchestrator: Orchestrator) -> None:
        """Wait for the orchestrator instances to run

        In the case where the orchestrator is launched as a batch
        through a WLM, we wait for the orchestrator to exit the
        queue before proceeding so new launched entities can
        be launched with SSDB address

        :param orchestrator: orchestrator instance
        :raises SmartSimError: if launch fails or manually stopped by user
        """
        if orchestrator.batch:
            logger.info("Orchestrator launched as a batch")
            logger.info("While queued, SmartSim will wait for Orchestrator to run")
            logger.info("CTRL+C interrupt to abort and cancel launch")

        ready = False
        while not ready:
            try:
                time.sleep(CONFIG.jm_interval)
                # manually trigger job update if JM not running
                if not self._jobs.actively_monitoring:
                    self._jobs.check_jobs()

                # _jobs.get_status acquires JM lock for main thread, no need for locking
                statuses = self.get_entity_list_status(orchestrator)
                if all(stat == SmartSimStatus.STATUS_RUNNING for stat in statuses):
                    ready = True
                    # TODO: Add a node status check
                elif any(stat in TERMINAL_STATUSES for stat in statuses):
                    self.stop_db(orchestrator)
                    msg = "Orchestrator failed during startup"
                    msg += f" See {orchestrator.path} for details"
                    raise SmartSimError(msg)
                else:
                    logger.debug("Waiting for orchestrator instances to spin up...")
            except KeyboardInterrupt:
                logger.info("Orchestrator launch cancelled - requesting to stop")
                self.stop_db(orchestrator)

                # re-raise keyboard interrupt so the job manager will display
                # any running and un-killed jobs as this method is only called
                # during launch and we handle all keyboard interrupts during
                # launch explicitly
                raise

    def reload_saved_db(
        self, checkpoint_file: t.Union[str, os.PathLike[str]]
    ) -> Orchestrator:
        with JM_LOCK:

            if not osp.exists(checkpoint_file):
                raise FileNotFoundError(
                    f"The SmartSim database config file {os.fspath(checkpoint_file)} "
                    "cannot be found."
                )

            try:
                with open(checkpoint_file, "rb") as pickle_file:
                    db_config = pickle.load(pickle_file)
            except (OSError, IOError) as e:
                msg = "Database checkpoint corrupted"
                raise SmartSimError(msg) from e

            err_message = (
                "The SmartSim database checkpoint is incomplete or corrupted. "
            )
            if not "db" in db_config:
                raise SmartSimError(
                    err_message + "Could not find the orchestrator object."
                )

            if not "db_jobs" in db_config:
                raise SmartSimError(
                    err_message + "Could not find database job objects."
                )

            if not "steps" in db_config:
                raise SmartSimError(
                    err_message + "Could not find database job objects."
                )
            orc: Orchestrator = db_config["db"]

            # TODO check that each db_object is running

            job_steps = zip(db_config["db_jobs"].values(), db_config["steps"])
            try:
                for db_job, step in job_steps:
                    self._jobs.db_jobs[db_job.ename] = db_job
                    self._launcher.add_step_to_mapping_table(db_job.name, step)
                    if step.task_id:
                        self._launcher.task_manager.add_existing(int(step.task_id))
            except LauncherError as e:
                raise SmartSimError("Failed to reconnect orchestrator") from e

            # start job manager if not already started
            if not self._jobs.actively_monitoring:
                self._jobs.start()

            return orc

    def _set_dbobjects(self, manifest: Manifest) -> None:
        if not manifest.has_db_objects:
            return

        address_dict = self._jobs.get_db_host_addresses()
        for (
            db_id,
            db_addresses,
        ) in address_dict.items():
            db_name, name = unpack_db_identifier(db_id, "_")

            hosts = list({address.split(":")[0] for address in db_addresses})
            ports = list({int(address.split(":")[-1]) for address in db_addresses})

            if not db_is_active(hosts=hosts, ports=ports, num_shards=len(db_addresses)):
                raise SSInternalError("Cannot set DB Objects, DB is not running")

            os.environ[f"SSDB{db_name}"] = db_addresses[0]

            os.environ[f"SR_DB_TYPE{db_name}"] = (
                CLUSTERED if len(db_addresses) > 1 else STANDALONE
            )

            options = ConfigOptions.create_from_environment(name)
            client = Client(options, logger_name="SmartSim")

            for model in manifest.models:
                if not model.colocated:
                    for db_model in model.db_models:
                        set_ml_model(db_model, client)
                    for db_script in model.db_scripts:
                        set_script(db_script, client)

            for ensemble in manifest.ensembles:
                for db_model in ensemble.db_models:
                    set_ml_model(db_model, client)
                for db_script in ensemble.db_scripts:
                    set_script(db_script, client)
                for entity in ensemble.models:
                    if not entity.colocated:
                        # Set models which could belong only
                        # to the entities and not to the ensemble
                        # but avoid duplicates
                        for db_model in entity.db_models:
                            if db_model not in ensemble.db_models:
                                set_ml_model(db_model, client)
                        for db_script in entity.db_scripts:
                            if db_script not in ensemble.db_scripts:
                                set_script(db_script, client)

    def _start_telemetry_monitor(self, exp_dir: str) -> None:
        """Spawns a telemetry monitor process to keep track of the life times
        of the processes launched through this controller.

        :param exp_dir: An experiment directory
        """
        if (
            self._telemetry_monitor is None
            or self._telemetry_monitor.returncode is not None
        ):
            logger.debug("Starting telemetry monitor process")
            cmd = [
                sys.executable,
                "-m",
                "smartsim._core.entrypoints.telemetrymonitor",
                "-exp_dir",
                exp_dir,
                "-frequency",
                str(CONFIG.telemetry_frequency),
                "-cooldown",
                str(CONFIG.telemetry_cooldown),
            ]
            # pylint: disable-next=consider-using-with
            self._telemetry_monitor = subprocess.Popen(
                cmd,
                stderr=sys.stderr,
                stdout=sys.stdout,
                cwd=str(pathlib.Path(__file__).parent.parent.parent),
                shell=False,
            )



import argparse
import importlib.util
import os
import subprocess as sp
import sys
import typing as t

import smartsim.log
from smartsim._core._cli.utils import SMART_LOGGER_FORMAT, MenuItemConfig
from smartsim.error.errors import SmartSimCLIActionCancelled

_LOGGER = smartsim.log.get_logger("Smart", fmt=SMART_LOGGER_FORMAT)


def dynamic_execute(
    cmd: str, plugin_name: str
) -> t.Callable[[argparse.Namespace, t.List[str]], int]:
    def process_execute(
        _args: argparse.Namespace, unparsed_args: t.List[str], /
    ) -> int:
        try:
            spec = importlib.util.find_spec(cmd)
            if spec is None:
                raise AttributeError
        except (ModuleNotFoundError, AttributeError):
            _LOGGER.error(f"{cmd} plugin not found. Please ensure it is installed")
            return os.EX_CONFIG

        combined_cmd = [sys.executable, "-m", cmd] + unparsed_args

        try:
            completed_proc = sp.run(combined_cmd, check=False)
        except KeyboardInterrupt as ex:
            msg = f"{plugin_name} terminated by user"
            raise SmartSimCLIActionCancelled(msg) from ex
        return completed_proc.returncode

    return process_execute


def dashboard() -> MenuItemConfig:
    return MenuItemConfig(
        "dashboard",
        (
            "Start the SmartSim dashboard to monitor experiment output from a "
            "graphical user interface. This requires that the SmartSim Dashboard "
            "Package be installed. For more infromation please visit "
            "https://github.com/CrayLabs/SmartDashboard"
        ),
        dynamic_execute("smartdashboard", "Dashboard"),
        is_plugin=True,
    )


plugins = (dashboard,)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import argparse
import asyncio
import logging
import os
import os.path
import pathlib
import signal
import sys
import typing as t
from types import FrameType

import smartsim._core.config as cfg
from smartsim._core.utils.telemetry.telemetry import (
    TelemetryMonitor,
    TelemetryMonitorArgs,
)
from smartsim.log import DEFAULT_LOG_FORMAT, HostnameFilter

"""Telemetry Monitor entrypoint
Starts a long-running, standalone process that hosts a `TelemetryMonitor`"""


logger = logging.getLogger("TelemetryMonitor")


def register_signal_handlers(
    handle_signal: t.Callable[[int, t.Optional[FrameType]], None]
) -> None:
    """Register a signal handling function for all termination events

    :param handle_signal: the function to execute when a term signal is received
    """
    # NOTE: omitting kill because it is not catchable
    term_signals = [signal.SIGINT, signal.SIGQUIT, signal.SIGTERM, signal.SIGABRT]
    for signal_num in term_signals:
        signal.signal(signal_num, handle_signal)


def get_parser() -> argparse.ArgumentParser:
    """Instantiate a parser to process command line arguments

    :returns: An argument parser ready to accept required telemetry monitor parameters
    """
    arg_parser = argparse.ArgumentParser(description="SmartSim Telemetry Monitor")
    arg_parser.add_argument(
        "-exp_dir",
        type=str,
        help="Experiment root directory",
        required=True,
    )
    arg_parser.add_argument(
        "-frequency",
        type=float,
        help="Frequency of telemetry updates (in seconds))",
        required=True,
    )
    arg_parser.add_argument(
        "-cooldown",
        type=int,
        help="Default lifetime of telemetry monitor (in seconds) before auto-shutdown",
        default=cfg.CONFIG.telemetry_cooldown,
    )
    arg_parser.add_argument(
        "-loglevel",
        type=int,
        help="Logging level",
        default=logging.INFO,
    )
    return arg_parser


def parse_arguments() -> TelemetryMonitorArgs:
    """Parse the command line arguments and return an instance
    of TelemetryMonitorArgs populated with the CLI inputs

    :returns: `TelemetryMonitorArgs` instance populated with command line arguments
    """
    parser = get_parser()
    parsed_args = parser.parse_args()
    return TelemetryMonitorArgs(
        parsed_args.exp_dir,
        parsed_args.frequency,
        parsed_args.cooldown,
        parsed_args.loglevel,
    )


def configure_logger(logger_: logging.Logger, log_level_: int, exp_dir: str) -> None:
    """Configure the telemetry monitor logger to write logs to the
    target output file path passed as an argument to the entrypoint

    :param logger_: logger to configure
    :param log_level_: log level to apply to the python logging system
    :param exp_dir: root path to experiment outputs
    """
    logger_.setLevel(log_level_)
    logger_.propagate = False

    # use a standard subdirectory of the experiment output path for logs
    telemetry_dir = pathlib.Path(exp_dir) / cfg.CONFIG.telemetry_subdir

    # all telemetry monitor logs are written to file in addition to stdout
    log_path = telemetry_dir / "logs/telemetrymonitor.out"
    log_path.parent.mkdir(parents=True, exist_ok=True)
    file_handler = logging.FileHandler(log_path, "a")

    # HostnameFilter is required to enrich log context to use DEFAULT_LOG_FORMAT
    file_handler.addFilter(HostnameFilter())

    formatter = logging.Formatter(DEFAULT_LOG_FORMAT)
    file_handler.setFormatter(formatter)
    logger_.addHandler(file_handler)


if __name__ == "__main__":
    """Prepare the telemetry monitor process using command line arguments.

    Sample usage:
    python -m smartsim._core.entrypoints.telemetrymonitor -exp_dir <exp_dir>
          -frequency 30 -cooldown 90 -loglevel INFO
    The experiment id is generated during experiment startup
    and can be found in the manifest.json in <exp_dir>/.smartsim/telemetry
    """
    os.environ["PYTHONUNBUFFERED"] = "1"

    args = parse_arguments()
    configure_logger(logger, args.log_level, args.exp_dir)

    telemetry_monitor = TelemetryMonitor(args)

    # Must register cleanup before the main loop is running
    def cleanup_telemetry_monitor(_signo: int, _frame: t.Optional[FrameType]) -> None:
        """Create an enclosure on `manifest_observer` to avoid global variables"""
        logger.info("Shutdown signal received by telemetry monitor entrypoint")
        telemetry_monitor.cleanup()

    register_signal_handlers(cleanup_telemetry_monitor)

    try:
        asyncio.run(telemetry_monitor.run())
        sys.exit(0)
    except Exception:
        logger.exception(
            "Shutting down telemetry monitor due to unexpected error", exc_info=True
        )

    sys.exit(1)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import dataclasses
import pathlib
import stat
import typing as t
from enum import IntEnum

import zmq
import zmq.auth

from smartsim._core.config.config import Config
from smartsim.log import get_logger

logger = get_logger(__name__)


class _KeyPermissions(IntEnum):
    """Permissions used by KeyManager"""

    PRIVATE_KEY = stat.S_IRUSR | stat.S_IWUSR
    """Permissions only allowing an owner to read and write the file"""
    PUBLIC_KEY = stat.S_IRUSR | stat.S_IWUSR | stat.S_IROTH | stat.S_IRGRP
    """Permissions allowing an owner, others, and the group to read a file"""

    PRIVATE_DIR = (
        stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IXOTH | stat.S_IXGRP
    )
    """Permissions allowing only owners to read, write and traverse a directory"""
    PUBLIC_DIR = (
        stat.S_IRUSR
        | stat.S_IWUSR
        | stat.S_IXUSR
        | stat.S_IROTH
        | stat.S_IXOTH
        | stat.S_IRGRP
        | stat.S_IXGRP
    )
    """Permissions allowing non-owners to traverse a directory"""


@dataclasses.dataclass(frozen=True)
class KeyPair:
    """A public and private key pair"""

    public: bytes = dataclasses.field(default=b"")
    """The public key"""

    private: bytes = dataclasses.field(default=b"", repr=False)
    """The private key"""

    @property
    def empty(self) -> bool:
        """Return `True` if the KeyPair has no key values set. Useful
        for faking the null object pattern"""
        return self.public == self.private and len(self.public) == 0


class _KeyLocator:
    """Determines the paths to use when persisting a `KeyPair` to disk"""

    def __init__(
        self,
        root_dir: pathlib.Path,
        filename: str,
        category: str,
    ) -> None:
        """Initiailize a `KeyLocator`

        :param root_dir: root path where keys are persisted to disk
        :param filename: the stem name of the key file
        :param category: the category or use-case for the key (e.g. server)
        :param separate_keys: flag indicating if public and private keys should
        be persisted in separate, corresponding directories
        """

        # constants for standardized paths.
        self._public_subdir = "pub"
        """The category subdirectory to use when persisting a public key"""

        self._private_subdir = "priv"
        """The category subdirectory to use when persisting a private key"""

        self._public_extension = "key"
        """The extension found on public keys"""

        self._private_extension = "key_secret"
        """The extension found on private keys"""

        self._key_root_dir = root_dir
        """Path to the root directory containing key files"""

        self._filename = filename
        """Base name for key files"""

        self._category = category
        """Category name used to further separate key locations"""

    @property
    def public_dir(self) -> pathlib.Path:
        """Target directory for the public key"""
        return self.public.parent

    @property
    def private_dir(self) -> pathlib.Path:
        """Target directory for the private key"""
        return self.private.parent

    @property
    def public_filename(self) -> str:
        """Filename (<stem>.<suffix>) of the public key file"""
        return f"{self._filename}.{self._public_extension}"

    @property
    def private_filename(self) -> str:
        """Filename (<stem>.<suffix>) of the private key file"""
        return f"{self._filename}.{self._private_extension}"

    @property
    def public(self) -> pathlib.Path:
        """Full target path of the public key file"""
        # combine the root and key type (e.g. /foo/bar + /server)
        # then combine the pub/priv key subdir (e.g. /foo/bar/server + /pub)
        path = self._key_root_dir / self._category / self._public_subdir
        return path / self.public_filename

    @property
    def private(self) -> pathlib.Path:
        """Full target path of the private key file"""
        # combine the root and key type (e.g. /foo/bar + /server)
        # then combine the pub/priv key subdir (e.g. /foo/bar/server + /pub)
        path = self._key_root_dir / self._category / self._private_subdir
        # combine the pub/priv key subdir if necessary (e.g. /foo/bar + /priv)

        return path / self.private_filename


class KeyManager:
    def __init__(
        self, config: Config, as_server: bool = False, as_client: bool = False
    ) -> None:
        """Initialize a KeyManager instance.
        :param config: SmartSim configuration
        :param as_server: flag to indicate when executing in the server context;
        set to `True` to avoid loading client secret key
        :param as_client: flag to indicate when executing in the client context;
        set to `True` to avoid loading server secret key
        """

        self._as_server = as_server
        """Set to `True` to return keys appropriate for the server context"""

        self._as_client = as_client
        """Set to `True` to return keys appropriate for the client context"""

        key_dir = pathlib.Path(config.smartsim_key_path).resolve()

        # Results in key path such as <key_root>/server/pub/smartsim.key
        self._server_locator = _KeyLocator(key_dir, "smartsim", "server")
        """The locator for producing the paths to store server key files"""

        # Results in key path such as <key_root>/client/pub/smartsim.key
        self._client_locator = _KeyLocator(key_dir, "smartsim", "client")
        """The locator for producing the paths to store client key files"""

    def create_directories(self) -> None:
        """Create the subdirectory structure necessary to hold
        the public and private key pairs for servers & clients"""
        for locator in [self._server_locator, self._client_locator]:
            if not locator.public_dir.exists():
                permission = _KeyPermissions.PUBLIC_DIR
                logger.debug(f"Creating key dir: {locator.public_dir}, {permission}")
                locator.public_dir.mkdir(parents=True, mode=permission)

            if not locator.private_dir.exists():
                permission = _KeyPermissions.PRIVATE_DIR
                logger.debug(f"Creating key dir: {locator.private_dir}, {permission}")
                locator.private_dir.mkdir(parents=True, mode=permission)

    @classmethod
    def _load_keypair(cls, locator: _KeyLocator, in_context: bool) -> KeyPair:
        """Load a specific `KeyPair` from disk

        :param locator: a `KeyLocator` that specifies the path to an existing key
        :param in_context: Boolean flag indicating if the keypair is the active
        context; ensures the public and private keys are both loaded when `True`.
        Only the public key is loaded when `False`
        :returns: a KeyPair containing the loaded public/private key
        """
        # private keys contain public & private key parts
        key_path = locator.private if in_context else locator.public

        pub_key: bytes = b""
        priv_key: t.Optional[bytes] = b""

        if key_path.exists():
            logger.debug(f"Existing key files located at {key_path}")
            pub_key, priv_key = zmq.auth.load_certificate(key_path)
        else:
            logger.debug(f"No key files found at {key_path}")

        # avoid a `None` value in the private key when it isn't loaded
        return KeyPair(pub_key, priv_key or b"")

    def _load_keys(self) -> t.Tuple[KeyPair, KeyPair]:
        """Use ZMQ auth to load public/private key pairs for the server and client
        components from the standard key paths for the associated experiment

        :returns: 2-tuple of `KeyPair` (server_keypair, client_keypair)
        ]"""
        try:
            server_keys = self._load_keypair(self._server_locator, self._as_server)
            client_keys = self._load_keypair(self._client_locator, self._as_client)

            return server_keys, client_keys
        except (ValueError, OSError):
            # expected if no keys could be loaded from disk
            logger.warning("Loading key pairs failed.", exc_info=True)

        return KeyPair(), KeyPair()

    @classmethod
    def _move_public_key(cls, locator: _KeyLocator) -> None:
        """The public and private key pair are created in the same directory. Move
        the public key out of the private subdir and into the public subdir

        :param locator: `KeyLocator` that determines the path to the
        key pair persisted in the same directory.
        """
        new_path = locator.private.with_suffix(locator.public.suffix)
        if new_path != locator.public:
            logger.debug(f"Moving key file from {locator.public} to {new_path}")
            new_path.rename(locator.public)

    def _create_keys(self) -> None:
        """Create and persist key files to disk"""
        for locator in [self._server_locator, self._client_locator]:
            # create keys in the private directory...
            zmq.auth.create_certificates(locator.private_dir, locator.private.stem)

            # ...but move the public key out of the private subdirectory
            self._move_public_key(locator)

            # and ensure correct r/w/x permissions on each file.
            locator.private.chmod(_KeyPermissions.PRIVATE_KEY)
            locator.public.chmod(_KeyPermissions.PUBLIC_KEY)

    def get_keys(self, create: bool = True) -> t.Tuple[KeyPair, KeyPair]:
        """Use ZMQ auth to generate a public/private key pair for the server
        and client components.

        :param no_create: pass `no_create=True` to ensure keys are not
        created and only pre-existing keys can be loaded
        :returns: 2-tuple of `KeyPair` (server_keypair, client_keypair)
        """
        logger.debug(f"Loading keys, creation {'is' if create else 'not'} allowed")
        server_keys, client_keys = self._load_keys()

        # check if we received "empty keys"
        if not server_keys.empty or not client_keys.empty:
            return server_keys, client_keys

        if not create:
            # if directed not to create new keys, return "empty keys"
            logger.debug("Returning empty key pairs")
            return KeyPair(), KeyPair()

        self.create_directories()
        self._create_keys()

        # load keys to ensure they were persisted
        return self._load_keys()

    @property
    def client_keys_dir(self) -> pathlib.Path:
        "Return the path to the client public keys directory"
        return self._client_locator.public_dir



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import itertools
import re
import sys
import typing as t
import warnings
from os import getcwd
from os import path as osp

from .._core._install.builder import Device
from .._core.utils.helpers import cat_arg_and_value
from ..error import EntityExistsError, SSUnsupportedError
from ..log import get_logger
from ..settings.base import BatchSettings, RunSettings
from .dbobject import DBModel, DBScript
from .entity import SmartSimEntity
from .files import EntityFiles

logger = get_logger(__name__)


class Model(SmartSimEntity):
    def __init__(
        self,
        name: str,
        params: t.Dict[str, str],
        run_settings: RunSettings,
        path: t.Optional[str] = getcwd(),
        params_as_args: t.Optional[t.List[str]] = None,
        batch_settings: t.Optional[BatchSettings] = None,
    ):
        """Initialize a ``Model``

        :param name: name of the model
        :param params: model parameters for writing into configuration files or
                       to be passed as command line arguments to executable.
        :param path: path to output, error, and configuration files
        :param run_settings: launcher settings specified in the experiment
        :param params_as_args: list of parameters which have to be
                               interpreted as command line arguments to
                               be added to run_settings
        :param batch_settings: Launcher settings for running the individual
                               model as a batch job
        """
        super().__init__(name, str(path), run_settings)
        self.params = params
        self.params_as_args = params_as_args
        self.incoming_entities: t.List[SmartSimEntity] = []
        self._key_prefixing_enabled = False
        self.batch_settings = batch_settings
        self._db_models: t.List[DBModel] = []
        self._db_scripts: t.List[DBScript] = []
        self.files: t.Optional[EntityFiles] = None

    @property
    def db_models(self) -> t.Iterable[DBModel]:
        """Retrieve an immutable collection of attached models

        :return: Return an immutable collection of attached models
        """
        return (model for model in self._db_models)

    @property
    def db_scripts(self) -> t.Iterable[DBScript]:
        """Retrieve an immutable collection attached of scripts

        :return: Return an immutable collection of attached scripts
        """
        return (script for script in self._db_scripts)

    @property
    def colocated(self) -> bool:
        """Return True if this Model will run with a colocated Orchestrator

        :return: Return True of the Model will run with a colocated Orchestrator
        """
        return bool(self.run_settings.colocated_db_settings)

    def register_incoming_entity(self, incoming_entity: SmartSimEntity) -> None:
        """Register future communication between entities.

        Registers the named data sources that this entity
        has access to by storing the key_prefix associated
        with that entity

        :param incoming_entity: The entity that data will be received from
        :raises SmartSimError: if incoming entity has already been registered
        """
        if incoming_entity.name in [
            in_entity.name for in_entity in self.incoming_entities
        ]:
            raise EntityExistsError(
                f"'{incoming_entity.name}' has already "
                + "been registered as an incoming entity"
            )

        self.incoming_entities.append(incoming_entity)

    def enable_key_prefixing(self) -> None:
        """If called, the entity will prefix its keys with its own model name"""
        self._key_prefixing_enabled = True

    def disable_key_prefixing(self) -> None:
        """If called, the entity will not prefix its keys with its own model name"""
        self._key_prefixing_enabled = False

    def query_key_prefixing(self) -> bool:
        """Inquire as to whether this entity will prefix its keys with its name

        :return: Return True if entity will prefix its keys with its name
        """
        return self._key_prefixing_enabled

    def attach_generator_files(
        self,
        to_copy: t.Optional[t.List[str]] = None,
        to_symlink: t.Optional[t.List[str]] = None,
        to_configure: t.Optional[t.List[str]] = None,
    ) -> None:
        """Attach files to an entity for generation

        Attach files needed for the entity that, upon generation,
        will be located in the path of the entity.  Invoking this method
        after files have already been attached will overwrite
        the previous list of entity files.

        During generation, files "to_copy" are copied into
        the path of the entity, and files "to_symlink" are
        symlinked into the path of the entity.

        Files "to_configure" are text based model input files where
        parameters for the model are set. Note that only models
        support the "to_configure" field. These files must have
        fields tagged that correspond to the values the user
        would like to change. The tag is settable but defaults
        to a semicolon e.g. THERMO = ;10;

        :param to_copy: files to copy
        :param to_symlink: files to symlink
        :param to_configure: input files with tagged parameters
        """
        to_copy = to_copy or []
        to_symlink = to_symlink or []
        to_configure = to_configure or []

        # Check that no file collides with the parameter file written
        # by Generator. We check the basename, even though it is more
        # restrictive than what we need (but it avoids relative path issues)
        for strategy in [to_copy, to_symlink, to_configure]:
            if strategy is not None and any(
                osp.basename(filename) == "smartsim_params.txt" for filename in strategy
            ):
                raise ValueError(
                    "`smartsim_params.txt` is a file automatically "
                    + "generated by SmartSim and cannot be ovewritten."
                )

        self.files = EntityFiles(to_configure, to_copy, to_symlink)

    @property
    def attached_files_table(self) -> str:
        """Return a list of attached files as a plain text table

        :returns: String version of table
        """
        if not self.files:
            return "No file attached to this model."
        return str(self.files)

    def print_attached_files(self) -> None:
        """Print a table of the attached files on std out"""
        print(self.attached_files_table)

    def colocate_db(self, *args: t.Any, **kwargs: t.Any) -> None:
        """An alias for ``Model.colocate_db_tcp``"""
        warnings.warn(
            (
                "`colocate_db` has been deprecated and will be removed in a \n"
                "future release. Please use `colocate_db_tcp` or `colocate_db_uds`."
            ),
            FutureWarning,
        )
        self.colocate_db_tcp(*args, **kwargs)

    def colocate_db_uds(
        self,
        unix_socket: str = "/tmp/redis.socket",
        socket_permissions: int = 755,
        db_cpus: int = 1,
        custom_pinning: t.Optional[t.Iterable[t.Union[int, t.Iterable[int]]]] = None,
        debug: bool = False,
        db_identifier: str = "",
        **kwargs: t.Any,
    ) -> None:
        """Colocate an Orchestrator instance with this Model over UDS.

        This method will initialize settings which add an unsharded
        database to this Model instance. Only this Model will be able to communicate
        with this colocated database by using Unix Domain sockets.

        Extra parameters for the db can be passed through kwargs. This includes
        many performance, caching and inference settings.

        .. highlight:: python
        .. code-block:: python

            example_kwargs = {
                "maxclients": 100000,
                "threads_per_queue": 1,
                "inter_op_threads": 1,
                "intra_op_threads": 1,
                "server_threads": 2 # keydb only
            }

        Generally these don't need to be changed.

        :param unix_socket: path to where the socket file will be created
        :param socket_permissions: permissions for the socketfile
        :param db_cpus: number of cpus to use for orchestrator
        :param custom_pinning: CPUs to pin the orchestrator to. Passing an empty
                               iterable disables pinning
        :param debug: launch Model with extra debug information about the colocated db
        :param kwargs: additional keyword arguments to pass to the orchestrator database
        """

        if not re.match(r"^[a-zA-Z0-9.:\,_\-/]*$", unix_socket):
            raise ValueError(
                f"Invalid name for unix socket: {unix_socket}. Must only "
                "contain alphanumeric characters or . : _ - /"
            )
        uds_options: t.Dict[str, t.Union[int, str]] = {
            "unix_socket": unix_socket,
            "socket_permissions": socket_permissions,
            # This is hardcoded to 0 as recommended by redis for UDS
            "port": 0,
        }

        common_options = {
            "cpus": db_cpus,
            "custom_pinning": custom_pinning,
            "debug": debug,
            "db_identifier": db_identifier,
        }
        self._set_colocated_db_settings(uds_options, common_options, **kwargs)

    def colocate_db_tcp(
        self,
        port: int = 6379,
        ifname: t.Union[str, list[str]] = "lo",
        db_cpus: int = 1,
        custom_pinning: t.Optional[t.Iterable[t.Union[int, t.Iterable[int]]]] = None,
        debug: bool = False,
        db_identifier: str = "",
        **kwargs: t.Any,
    ) -> None:
        """Colocate an Orchestrator instance with this Model over TCP/IP.

        This method will initialize settings which add an unsharded
        database to this Model instance. Only this Model will be able to communicate
        with this colocated database by using the loopback TCP interface.

        Extra parameters for the db can be passed through kwargs. This includes
        many performance, caching and inference settings.

        .. highlight:: python
        .. code-block:: python

            ex. kwargs = {
                maxclients: 100000,
                threads_per_queue: 1,
                inter_op_threads: 1,
                intra_op_threads: 1,
                server_threads: 2 # keydb only
            }

        Generally these don't need to be changed.

        :param port: port to use for orchestrator database
        :param ifname: interface to use for orchestrator
        :param db_cpus: number of cpus to use for orchestrator
        :param custom_pinning: CPUs to pin the orchestrator to. Passing an empty
                               iterable disables pinning
        :param debug: launch Model with extra debug information about the colocated db
        :param kwargs: additional keyword arguments to pass to the orchestrator database
        """

        tcp_options = {"port": port, "ifname": ifname}
        common_options = {
            "cpus": db_cpus,
            "custom_pinning": custom_pinning,
            "debug": debug,
            "db_identifier": db_identifier,
        }
        self._set_colocated_db_settings(tcp_options, common_options, **kwargs)

    def _set_colocated_db_settings(
        self,
        connection_options: t.Mapping[str, t.Union[int, t.List[str], str]],
        common_options: t.Dict[
            str,
            t.Union[
                t.Union[t.Iterable[t.Union[int, t.Iterable[int]]], None],
                bool,
                int,
                str,
                None,
            ],
        ],
        **kwargs: t.Union[int, None],
    ) -> None:
        """
        Ingest the connection-specific options (UDS/TCP) and set the final settings
        for the colocated database
        """

        if hasattr(self.run_settings, "mpmd") and len(self.run_settings.mpmd) > 0:
            raise SSUnsupportedError(
                "Models colocated with databases cannot be run as a mpmd workload"
            )

        if hasattr(self.run_settings, "_prep_colocated_db"):
            # pylint: disable-next=protected-access
            self.run_settings._prep_colocated_db(common_options["cpus"])

        if "limit_app_cpus" in kwargs:
            raise SSUnsupportedError(
                "Pinning app CPUs via limit_app_cpus is not supported. Modify "
                "RunSettings using the correct binding option for your launcher."
            )

        # TODO list which db settings can be extras
        custom_pinning_ = t.cast(
            t.Optional[t.Iterable[t.Union[int, t.Iterable[int]]]],
            common_options.get("custom_pinning"),
        )
        cpus_ = t.cast(int, common_options.get("cpus"))
        common_options["custom_pinning"] = self._create_pinning_string(
            custom_pinning_, cpus_
        )

        colo_db_config: t.Dict[
            str,
            t.Union[
                bool,
                int,
                str,
                None,
                t.List[str],
                t.Iterable[t.Union[int, t.Iterable[int]]],
                t.List[DBModel],
                t.List[DBScript],
                t.Dict[str, t.Union[int, None]],
                t.Dict[str, str],
            ],
        ] = {}
        colo_db_config.update(connection_options)
        colo_db_config.update(common_options)

        redis_ai_temp = {
            "threads_per_queue": kwargs.get("threads_per_queue", None),
            "inter_op_parallelism": kwargs.get("inter_op_parallelism", None),
            "intra_op_parallelism": kwargs.get("intra_op_parallelism", None),
        }
        # redisai arguments for inference settings
        colo_db_config["rai_args"] = redis_ai_temp
        colo_db_config["extra_db_args"] = {
            k: str(v) for k, v in kwargs.items() if k not in redis_ai_temp
        }

        self._check_db_objects_colo()
        colo_db_config["db_models"] = self._db_models
        colo_db_config["db_scripts"] = self._db_scripts

        self.run_settings.colocated_db_settings = colo_db_config

    @staticmethod
    def _create_pinning_string(
        pin_ids: t.Optional[t.Iterable[t.Union[int, t.Iterable[int]]]], cpus: int
    ) -> t.Optional[str]:
        """Create a comma-separated string of CPU ids. By default, ``None``
        returns 0,1,...,cpus-1; an empty iterable will disable pinning
        altogether, and an iterable constructs a comma separated string of
        integers (e.g. ``[0, 2, 5]`` -> ``"0,2,5"``)
        """

        def _stringify_id(_id: int) -> str:
            """Return the cPU id as a string if an int, otherwise raise a ValueError"""
            if isinstance(_id, int):
                if _id < 0:
                    raise ValueError("CPU id must be a nonnegative number")
                return str(_id)

            raise TypeError(f"Argument is of type '{type(_id)}' not 'int'")

        try:
            pin_ids = tuple(pin_ids) if pin_ids is not None else None
        except TypeError:
            raise TypeError(
                "Expected a cpu pinning specification of type iterable of ints or "
                f"iterables of ints. Instead got type `{type(pin_ids)}`"
            ) from None

        # Deal with MacOSX limitations first. The "None" (default) disables pinning
        # and is equivalent to []. The only invalid option is a non-empty pinning
        if sys.platform == "darwin":
            if pin_ids:
                warnings.warn(
                    "CPU pinning is not supported on MacOSX. Ignoring pinning "
                    "specification.",
                    RuntimeWarning,
                )
            return None

        # Flatten the iterable into a list and check to make sure that the resulting
        # elements are all ints
        if pin_ids is None:
            return ",".join(_stringify_id(i) for i in range(cpus))
        if not pin_ids:
            return None
        pin_ids = ((x,) if isinstance(x, int) else x for x in pin_ids)
        to_fmt = itertools.chain.from_iterable(pin_ids)
        return ",".join(sorted({_stringify_id(x) for x in to_fmt}))

    def params_to_args(self) -> None:
        """Convert parameters to command line arguments and update run settings."""
        if self.params_as_args is not None:
            for param in self.params_as_args:
                if not param in self.params:
                    raise ValueError(
                        f"Tried to convert {param} to command line argument for Model "
                        f"{self.name}, but its value was not found in model params"
                    )
                if self.run_settings is None:
                    raise ValueError(
                        "Tried to configure command line parameter for Model "
                        f"{self.name}, but no RunSettings are set."
                    )
                self.run_settings.add_exe_args(
                    cat_arg_and_value(param, self.params[param])
                )

    def add_ml_model(
        self,
        name: str,
        backend: str,
        model: t.Optional[bytes] = None,
        model_path: t.Optional[str] = None,
        device: str = Device.CPU.value.upper(),
        devices_per_node: int = 1,
        first_device: int = 0,
        batch_size: int = 0,
        min_batch_size: int = 0,
        min_batch_timeout: int = 0,
        tag: str = "",
        inputs: t.Optional[t.List[str]] = None,
        outputs: t.Optional[t.List[str]] = None,
    ) -> None:
        """A TF, TF-lite, PT, or ONNX model to load into the DB at runtime

        Each ML Model added will be loaded into an
        orchestrator (converged or not) prior to the execution
        of this Model instance

        One of either model (in memory representation) or model_path (file)
        must be provided

        :param name: key to store model under
        :param backend: name of the backend (TORCH, TF, TFLITE, ONNX)
        :param model: A model in memory (only supported for non-colocated orchestrators)
        :param model_path: serialized model
        :param device: name of device for execution
        :param devices_per_node: The number of GPU devices available on the host.
               This parameter only applies to GPU devices and will be ignored if device
               is specified as CPU.
        :param first_device: The first GPU device to use on the host.
               This parameter only applies to GPU devices and will be ignored if device
               is specified as CPU.
        :param batch_size: batch size for execution
        :param min_batch_size: minimum batch size for model execution
        :param min_batch_timeout: time to wait for minimum batch size
        :param tag: additional tag for model information
        :param inputs: model inputs (TF only)
        :param outputs: model outupts (TF only)
        """
        db_model = DBModel(
            name=name,
            backend=backend,
            model=model,
            model_file=model_path,
            device=device,
            devices_per_node=devices_per_node,
            first_device=first_device,
            batch_size=batch_size,
            min_batch_size=min_batch_size,
            min_batch_timeout=min_batch_timeout,
            tag=tag,
            inputs=inputs,
            outputs=outputs,
        )
        self.add_ml_model_object(db_model)

    def add_script(
        self,
        name: str,
        script: t.Optional[str] = None,
        script_path: t.Optional[str] = None,
        device: str = Device.CPU.value.upper(),
        devices_per_node: int = 1,
        first_device: int = 0,
    ) -> None:
        """TorchScript to launch with this Model instance

        Each script added to the model will be loaded into an
        orchestrator (converged or not) prior to the execution
        of this Model instance

        Device selection is either "GPU" or "CPU". If many devices are
        present, a number can be passed for specification e.g. "GPU:1".

        Setting ``devices_per_node=N``, with N greater than one will result
        in the script being stored in the first N devices of type ``device``;
        alternatively, setting ``first_device=M`` will result in the script
        being stored on nodes M through M + N - 1.

        One of either script (in memory string representation) or script_path (file)
        must be provided

        :param name: key to store script under
        :param script: TorchScript code (only supported for non-colocated orchestrators)
        :param script_path: path to TorchScript code
        :param device: device for script execution
        :param devices_per_node: The number of GPU devices available on the host.
               This parameter only applies to GPU devices and will be ignored if device
               is specified as CPU.
        :param first_device: The first GPU device to use on the host.
               This parameter only applies to GPU devices and will be ignored if device
               is specified as CPU.
        """
        db_script = DBScript(
            name=name,
            script=script,
            script_path=script_path,
            device=device,
            devices_per_node=devices_per_node,
            first_device=first_device,
        )
        self.add_script_object(db_script)

    def add_function(
        self,
        name: str,
        function: t.Optional[str] = None,
        device: str = Device.CPU.value.upper(),
        devices_per_node: int = 1,
        first_device: int = 0,
    ) -> None:
        """TorchScript function to launch with this Model instance

        Each script function to the model will be loaded into a
        non-converged orchestrator prior to the execution
        of this Model instance.

        For converged orchestrators, the :meth:`add_script` method should be used.

        Device selection is either "GPU" or "CPU". If many devices are
        present, a number can be passed for specification e.g. "GPU:1".

        Setting ``devices_per_node=N``, with N greater than one will result
        in the model being stored in the first N devices of type ``device``.

        :param name: key to store function under
        :param function: TorchScript function code
        :param device: device for script execution
        :param devices_per_node: The number of GPU devices available on the host.
               This parameter only applies to GPU devices and will be ignored if device
               is specified as CPU.
        :param first_device: The first GPU device to use on the host.
               This parameter only applies to GPU devices and will be ignored if device
               is specified as CPU.
        """
        db_script = DBScript(
            name=name,
            script=function,
            device=device,
            devices_per_node=devices_per_node,
            first_device=first_device,
        )
        self.add_script_object(db_script)

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Model):
            return False

        if self.name == other.name:
            return True
        return False

    def __str__(self) -> str:  # pragma: no cover
        entity_str = "Name: " + self.name + "\n"
        entity_str += "Type: " + self.type + "\n"
        entity_str += str(self.run_settings) + "\n"
        if self._db_models:
            entity_str += "DB Models: \n" + str(len(self._db_models)) + "\n"
        if self._db_scripts:
            entity_str += "DB Scripts: \n" + str(len(self._db_scripts)) + "\n"
        return entity_str

    def add_ml_model_object(self, db_model: DBModel) -> None:
        if not db_model.is_file and self.colocated:
            err_msg = "ML model can not be set from memory for colocated databases.\n"
            err_msg += (
                f"Please store the ML model named {db_model.name} in binary format "
            )
            err_msg += "and add it to the SmartSim Model as file."
            raise SSUnsupportedError(err_msg)

        self._db_models.append(db_model)

    def add_script_object(self, db_script: DBScript) -> None:
        if db_script.func and self.colocated:
            if not isinstance(db_script.func, str):
                err_msg = (
                    "Functions can not be set from memory for colocated databases.\n"
                    f"Please convert the function named {db_script.name} "
                    "to a string or store it as a text file and add it to the "
                    "SmartSim Model with add_script."
                )
                raise SSUnsupportedError(err_msg)
        self._db_scripts.append(db_script)

    def _check_db_objects_colo(self) -> None:
        for db_model in self._db_models:
            if not db_model.is_file:
                err_msg = (
                    "ML model can not be set from memory for colocated databases.\n"
                    f"Please store the ML model named {db_model.name} in binary "
                    "format and add it to the SmartSim Model as file."
                )
                raise SSUnsupportedError(err_msg)

        for db_script in self._db_scripts:
            if db_script.func:
                if not isinstance(db_script.func, str):
                    err_msg = (
                        "Functions can not be set from memory for colocated "
                        "databases.\nPlease convert the function named "
                        f"{db_script.name} to a string or store it as a text"
                        "file and add it to the SmartSim Model with add_script."
                    )
                    raise SSUnsupportedError(err_msg)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import os
import shutil
import sys
import typing as t

from ...._core.schemas.dragonRequests import DragonRunRequest, request_registry
from ....error.errors import SSUnsupportedError
from ....log import get_logger
from ....settings import (
    DragonRunSettings,
    QsubBatchSettings,
    SbatchSettings,
    Singularity,
)
from .step import Step

logger = get_logger(__name__)


class DragonStep(Step):
    def __init__(self, name: str, cwd: str, run_settings: DragonRunSettings) -> None:
        """Initialize a srun job step

        :param name: name of the entity to be launched
        :param cwd: path to launch dir
        :param run_settings: run settings for entity
        """
        super().__init__(name, cwd, run_settings)
        self.managed = True

    @property
    def run_settings(self) -> DragonRunSettings:
        return t.cast(DragonRunSettings, self.step_settings)

    def get_launch_cmd(self) -> t.List[str]:
        """Get stringified version of request
         needed to launch this step

        :return: launch command
        """
        run_settings = self.run_settings
        exe_cmd = []

        if run_settings.colocated_db_settings:
            # Replace the command with the entrypoint wrapper script
            bash = shutil.which("bash")
            if not bash:
                raise RuntimeError("Could not find bash in PATH")
            launch_script_path = self.get_colocated_launch_script()
            exe_cmd += [bash, launch_script_path]

        if isinstance(run_settings.container, Singularity):
            # pylint: disable-next=protected-access
            exe_cmd += run_settings.container._container_cmds(self.cwd)

        exe_cmd += run_settings.exe

        exe_args = self._get_exe_args_list(run_settings)

        exe_cmd_and_args = exe_cmd + exe_args

        return exe_cmd_and_args

    @staticmethod
    def _get_exe_args_list(run_setting: DragonRunSettings) -> t.List[str]:
        """Convenience function to encapsulate checking the
        runsettings.exe_args type to always return a list
        """
        exe_args = run_setting.exe_args
        args: t.List[str] = exe_args if isinstance(exe_args, list) else [exe_args]
        return args


class DragonBatchStep(Step):
    def __init__(
        self,
        name: str,
        cwd: str,
        batch_settings: t.Union[SbatchSettings, QsubBatchSettings],
    ) -> None:
        """Initialize a Slurm Sbatch step

        :param name: name of the entity to launch
        :param cwd: path to launch dir
        :param batch_settings: batch settings for entity
        """
        super().__init__(name, cwd, batch_settings)
        self.steps: t.List[Step] = []
        self.managed = True
        self.batch_settings = batch_settings
        self._request_file_name = "requests.json"

    def get_launch_cmd(self) -> t.List[str]:
        """Get the launch command for the batch

        :return: launch command for the batch
        """
        if isinstance(self.batch_settings, SbatchSettings):
            script = self._write_sbatch_script()
            return [self.batch_settings.batch_cmd, "--parsable", script]
        if isinstance(self.batch_settings, QsubBatchSettings):
            script = self._write_qsub_script()
            return [self.batch_settings.batch_cmd, script]

        raise SSUnsupportedError(
            "DragonBatchStep only support SbatchSettings and QsubBatchSettings"
        )

    def add_to_batch(self, step: Step) -> None:
        """Add a job step to this batch

        :param step: a job step instance e.g. DragonStep
        """
        self.steps.append(step)
        logger.debug(f"Added step command to batch for {step.name}")

    @staticmethod
    def _dragon_entrypoint_cmd(request_file: str) -> str:
        """Return command needed to run the Dragon entrypoint"""
        cmd = [
            sys.executable,
            "-m",
            "smartsim._core.entrypoints.dragon_client",
            "+submit",
            request_file,
        ]
        return " ".join(cmd)

    def _write_request_file(self) -> str:
        """Write json file with requests to submit to Dragon server"""
        request_file = self.get_step_file(
            ending="json", script_name=self._request_file_name
        )
        requests = []
        for step in self.steps:
            run_settings = t.cast(DragonRunSettings, step.step_settings)
            run_args = run_settings.run_args
            env = run_settings.env_vars
            nodes = int(run_args.get("nodes", None) or 1)
            tasks_per_node = int(run_args.get("tasks-per-node", None) or 1)

            cmd = step.get_launch_cmd()
            out, err = step.get_output_files()
            request = DragonRunRequest(
                exe=cmd[0],
                exe_args=cmd[1:],
                path=step.cwd,
                name=step.name,
                nodes=nodes,
                tasks_per_node=tasks_per_node,
                env=env,
                current_env=os.environ,
                output_file=out,
                error_file=err,
            )
            requests.append(request_registry.to_string(request))
        with open(request_file, "w", encoding="utf-8") as script_file:
            script_file.write(json.dumps(requests))

        return request_file

    def _write_sbatch_script(self) -> str:
        """Write the PBS batch script

        :return: batch script path after writing
        """
        batch_script = self.get_step_file(ending=".sh")
        output, error = self.get_output_files()
        request_file = self._write_request_file()
        with open(batch_script, "w", encoding="utf-8") as script_file:
            script_file.write("#!/bin/bash\n\n")
            script_file.write(f"#SBATCH --output={output}\n")
            script_file.write(f"#SBATCH --error={error}\n")
            script_file.write(f"#SBATCH --job-name={self.name}\n")

            # add additional sbatch options
            for opt in self.batch_settings.format_batch_args():
                script_file.write(f"#SBATCH {opt}\n")

            script_file.write(
                f"#SBATCH --export=ALL,SMARTSIM_DRAGON_SERVER_PATH={self.cwd},"
                "PYTHONUNBUFFERED=1\n"
            )

            for cmd in self.batch_settings.preamble:
                script_file.write(f"{cmd}\n")

            script_file.write(
                DragonBatchStep._dragon_entrypoint_cmd(request_file) + "\n"
            )
        return batch_script

    def _write_qsub_script(self) -> str:
        """Write the Slurm batch script

        :return: batch script path after writing
        """
        batch_script = self.get_step_file(ending=".sh")
        output, error = self.get_output_files()
        request_file = self._write_request_file()
        with open(batch_script, "w", encoding="utf-8") as script_file:
            script_file.write("#!/bin/bash\n\n")
            script_file.write(f"#PBS -o {output}\n")
            script_file.write(f"#PBS -e {error}\n")
            script_file.write(f"#PBS -N {self.name}\n")
            script_file.write("#PBS -V \n")

            # add additional sbatch options
            for opt in self.batch_settings.format_batch_args():
                script_file.write(f"#PBS {opt}\n")

            script_file.write(f"#PBS -v SMARTSIM_DRAGON_SERVER_PATH={self.cwd}\n")

            for cmd in self.batch_settings.preamble:
                script_file.write(f"{cmd}\n")

            script_file.write(
                DragonBatchStep._dragon_entrypoint_cmd(request_file) + "\n"
            )

        return batch_script



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys

from smartsim._core._cli.cli import default_cli
from smartsim._core._cli.utils import SMART_LOGGER_FORMAT
from smartsim.error.errors import SmartSimCLIActionCancelled
from smartsim.log import get_logger

logger = get_logger("Smart", fmt=SMART_LOGGER_FORMAT)


def main() -> int:
    smart_cli = default_cli()
    exception_trace_back_msg = "SmartSim exited with the following exception info:"

    try:
        return smart_cli.execute(sys.argv)
    except SmartSimCLIActionCancelled as ssi:
        logger.info(str(ssi))
        logger.debug(exception_trace_back_msg, exc_info=ssi)
    except KeyboardInterrupt as e:
        logger.info("SmartSim was terminated by user")
        logger.debug(exception_trace_back_msg, exc_info=e)
    return os.EX_OK


if __name__ == "__main__":
    main()



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
import typing as t
from os import path

from tabulate import tabulate


class EntityFiles:
    """EntityFiles are the files a user wishes to have available to
    models and nodes within SmartSim. Each entity has a method
    `entity.attach_generator_files()` that creates one of these
    objects such that at generation time, each file type will be
    present within the generated model or node directory.

    Tagged files are the configuration files for a model that
    can be searched through and edited by the ModelWriter.

    Copy files are files that a user wants to copy into the
    model or node directory without searching through and
    editing them for tags.

    Lastly, symlink can be used for big datasets or input
    files that a user just wants to have present in the directory
    without necessary having to copy the entire file.
    """

    def __init__(
        self,
        tagged: t.Optional[t.List[str]] = None,
        copy: t.Optional[t.List[str]] = None,
        symlink: t.Optional[t.List[str]] = None,
    ) -> None:
        """Initialize an EntityFiles instance

        :param tagged: tagged files for model configuration
        :param copy: files or directories to copy into model
                     or node directories
        :param symlink: files to symlink into model or node
                        directories
        """
        self.tagged = tagged or []
        self.copy = copy or []
        self.link = symlink or []
        self.tagged_hierarchy = None
        self._check_files()

    def _check_files(self) -> None:
        """Ensure the files provided by the user are of the correct
           type and actually exist somewhere on the filesystem.

        :raises SSConfigError: If a user provides a directory within
                               the tagged files.
        """

        # type check all files provided by user
        self.tagged = self._type_check_files(self.tagged, "Tagged")
        self.copy = self._type_check_files(self.copy, "Copyable")
        self.link = self._type_check_files(self.link, "Symlink")

        self.tagged_hierarchy = TaggedFilesHierarchy.from_list_paths(
            self.tagged, dir_contents_to_base=True
        )

        for i, value in enumerate(self.copy):
            self.copy[i] = self._check_path(value)

        for i, value in enumerate(self.link):
            self.link[i] = self._check_path(value)

    @staticmethod
    def _type_check_files(
        file_list: t.Union[t.List[str], None], file_type: str
    ) -> t.List[str]:
        """Check the type of the files provided by the user.

        :param file_list: either tagged, copy, or symlink files
        :param file_type: name of the file type e.g. "tagged"
        :raises TypeError: if incorrect type is provided by user
        :return: file list provided
        """
        if file_list:
            if not isinstance(file_list, list):
                if isinstance(file_list, str):
                    file_list = [file_list]
                else:
                    raise TypeError(
                        f"{file_type} files given were not of type list or str"
                    )
            else:
                if not all(isinstance(f, str) for f in file_list):
                    raise TypeError(f"Not all {file_type} files were of type str")
        return file_list or []

    @staticmethod
    def _check_path(file_path: str) -> str:
        """Given a user provided path-like str, find the actual path to
           the directory or file and create a full path.

        :param file_path: path to a specific file or directory
        :raises FileNotFoundError: if file or directory does not exist
        :return: full path to file or directory
        """
        full_path = path.abspath(file_path)
        if path.isfile(full_path):
            return full_path
        if path.isdir(full_path):
            return full_path
        raise FileNotFoundError(f"File or Directory {file_path} not found")

    def __str__(self) -> str:
        """Return table summarizing attached files."""
        values = []

        if self.copy:
            values.append(["Copy", "\n".join(self.copy)])
        if self.link:
            values.append(["Symlink", "\n".join(self.link)])
        if self.tagged:
            values.append(["Configure", "\n".join(self.tagged)])

        if not values:
            return "No file attached to this entity."

        return tabulate(values, headers=["Strategy", "Files"], tablefmt="grid")


class TaggedFilesHierarchy:
    """The TaggedFilesHierarchy represents a directory
    containing potentially tagged files and subdirectories.

    TaggedFilesHierarchy.base is the directory path from
    the the root of the generated file structure

    TaggedFilesHierarchy.files is a collection of paths to
    files that need to be copied to directory that the
    TaggedFilesHierarchy represents

    TaggedFilesHierarchy.dirs is a collection of child
    TaggedFilesHierarchy, each representing a subdirectory
    that needs to generated

    By performing a depth first search over the entire
    hierarchy starting at the root directory structure, the
    tagged file directory structure can be replicated
    """

    def __init__(self, parent: t.Optional[t.Any] = None, subdir_name: str = "") -> None:
        """Initialize a TaggedFilesHierarchy

        :param parent: The parent hierarchy of the new hierarchy,
                       must be None if creating a root hierarchy,
                       must be provided if creating a subhierachy
        :param subdir_name: Name of subdirectory representd by the new hierarchy,
                            must be "" if creating a root hierarchy,
                            must be any valid dir name if subhierarchy,
                            invalid names are ".", ".." or contain path seperators
        :raises ValueError: if given a subdir_name without a parent,
                            if given a parent without a subdir_name,
                            or if the subdir_name is invalid
        """
        if parent is None and subdir_name:
            raise ValueError(
                "TaggedFilesHierarchies should not have a subdirectory name without a"
                + " parent"
            )
        if parent is not None and not subdir_name:
            raise ValueError(
                "Child TaggedFilesHierarchies must have a subdirectory name"
            )
        if subdir_name in {".", ".."} or path.sep in subdir_name:
            raise ValueError(
                "Child TaggedFilesHierarchies subdirectory names must not contain"
                + " path seperators or be reserved dirs '.' or '..'"
            )

        if parent:
            parent.dirs.add(self)

        self._base: str = path.join(parent.base, subdir_name) if parent else ""
        self.parent: t.Any = parent
        self.files: t.Set[str] = set()
        self.dirs: t.Set[TaggedFilesHierarchy] = set()

    @property
    def base(self) -> str:
        """Property to ensure that self.base is read-only"""
        return self._base

    @classmethod
    def from_list_paths(
        cls, path_list: t.List[str], dir_contents_to_base: bool = False
    ) -> t.Any:
        """Given a list of absolute paths to files and dirs, create and return
        a TaggedFilesHierarchy instance representing the file hierarchy of
        tagged files. All files in the path list will be placed in the base of
        the file hierarchy.

        :param path_list: list of absolute paths to tagged files or dirs
                          containing tagged files
        :param dir_contents_to_base: When a top level dir is encountered, if
                                     this value is truthy, files in the dir are
                                     put into the base hierarchy level.
                                     Otherwise, a new sub level is created for
                                     the dir
        :return: A built tagged file hierarchy for the given files
        """
        tagged_file_hierarchy = cls()
        if dir_contents_to_base:
            new_paths = []
            for tagged_path in path_list:
                if os.path.isdir(tagged_path):
                    new_paths += [
                        os.path.join(tagged_path, file)
                        for file in os.listdir(tagged_path)
                    ]
                else:
                    new_paths.append(tagged_path)
            path_list = new_paths
        tagged_file_hierarchy._add_paths(path_list)
        return tagged_file_hierarchy

    def _add_file(self, file: str) -> None:
        """Add a file to the current level in the file hierarchy

        :param file: absoute path to a file to add to the hierarchy
        """
        self.files.add(file)

    def _add_dir(self, dir_path: str) -> None:
        """Add a dir contianing tagged files by creating a new sub level in the
        tagged file hierarchy. All paths within the directroy are added to the
        the new level sub level tagged file hierarchy

        :param dir: absoute path to a dir to add to the hierarchy
        """
        tagged_file_hierarchy = TaggedFilesHierarchy(self, path.basename(dir_path))
        # pylint: disable-next=protected-access
        tagged_file_hierarchy._add_paths(
            [path.join(dir_path, file) for file in os.listdir(dir_path)]
        )

    def _add_paths(self, paths: t.List[str]) -> None:
        """Takes a list of paths and iterates over it, determining if each
        path is to a file or a dir and then appropriatly adding it to the
        TaggedFilesHierarchy.

        :param paths: list of paths to files or dirs to add to the hierarchy
        :raises ValueError: if link to dir is found
        :raises FileNotFoundError: if path does not exist
        """
        for candidate in paths:
            candidate = os.path.abspath(candidate)
            if os.path.isdir(candidate):
                if os.path.islink(candidate):
                    raise ValueError(
                        "Tagged directories and thier subdirectories cannot be links"
                        + " to prevent circular directory structures"
                    )
                self._add_dir(candidate)
            elif os.path.isfile(candidate):
                self._add_file(candidate)
            else:
                raise FileNotFoundError(f"File or Directory {candidate} not found")



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import dataclasses
import typing as t

import pydantic
import pydantic.dataclasses

if t.TYPE_CHECKING:
    from zmq.sugar.socket import Socket

_SchemaT = t.TypeVar("_SchemaT", bound=pydantic.BaseModel)
_SendT = t.TypeVar("_SendT", bound=pydantic.BaseModel)
_RecvT = t.TypeVar("_RecvT", bound=pydantic.BaseModel)

_DEFAULT_MSG_DELIM: t.Final[str] = "|"


@t.final
@pydantic.dataclasses.dataclass(frozen=True)
class _Message(t.Generic[_SchemaT]):
    payload: _SchemaT
    header: str = pydantic.Field(min_length=1)
    delimiter: str = pydantic.Field(min_length=1, default=_DEFAULT_MSG_DELIM)

    def __str__(self) -> str:
        return self.delimiter.join((self.header, self.payload.json()))

    @classmethod
    def from_str(
        cls,
        str_: str,
        payload_type: t.Type[_SchemaT],
        delimiter: str = _DEFAULT_MSG_DELIM,
    ) -> "_Message[_SchemaT]":
        header, payload = str_.split(delimiter, 1)
        return cls(payload_type.parse_raw(payload), header, delimiter)


class SchemaRegistry(t.Generic[_SchemaT]):
    def __init__(
        self, init_map: t.Optional[t.Mapping[str, t.Type[_SchemaT]]] = None
    ) -> None:
        self._map = dict(init_map) if init_map else {}

    def register(self, key: str) -> t.Callable[[t.Type[_SchemaT]], t.Type[_SchemaT]]:
        if _DEFAULT_MSG_DELIM in key:
            _msg = f"Registry key cannot contain delimiter `{_DEFAULT_MSG_DELIM}`"
            raise ValueError(_msg)
        if not key:
            raise KeyError("Key cannot be the empty string")
        if key in self._map:
            raise KeyError(f"Key `{key}` has already been registered for this parser")

        def _register(cls: t.Type[_SchemaT]) -> t.Type[_SchemaT]:
            self._map[key] = cls
            return cls

        return _register

    def to_string(self, schema: _SchemaT) -> str:
        return str(self._to_message(schema))

    def _to_message(self, schema: _SchemaT) -> _Message[_SchemaT]:
        reverse_map = dict((v, k) for k, v in self._map.items())
        try:
            val = reverse_map[type(schema)]
        except KeyError:
            raise TypeError(f"Unregistered schema type: {type(schema)}") from None
        return _Message(schema, val, _DEFAULT_MSG_DELIM)

    def from_string(self, str_: str) -> _SchemaT:
        try:
            type_, _ = str_.split(_DEFAULT_MSG_DELIM, 1)
        except ValueError:
            _msg = f"Failed to determine schema type of the string {repr(str_)}"
            raise ValueError(_msg) from None
        try:
            cls = self._map[type_]
        except KeyError:
            raise ValueError(f"No type of value `{type_}` is registered") from None
        msg = _Message.from_str(str_, cls, _DEFAULT_MSG_DELIM)
        return self._from_message(msg)

    @staticmethod
    def _from_message(msg: _Message[_SchemaT]) -> _SchemaT:
        return msg.payload


@dataclasses.dataclass(frozen=True)
class SocketSchemaTranslator(t.Generic[_SendT, _RecvT]):
    socket: "Socket[t.Any]"
    _send_registry: SchemaRegistry[_SendT]
    _recv_registry: SchemaRegistry[_RecvT]

    def send(self, schema: _SendT, flags: int = 0) -> None:
        self.socket.send_string(self._send_registry.to_string(schema), flags)

    def recv(self, flags: int = 0) -> _RecvT:
        return self._recv_registry.from_string(self.socket.recv_string(flags))



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import typing as t

# Exceptions


class SmartSimError(Exception):
    """Base SmartSim error"""


class SSUnsupportedError(Exception):
    """Raised in the event that a called method isn't supported by SmartSim yet"""


class EntityExistsError(SmartSimError):
    """Raised when a user tries to create an entity or files/directories for
    an entity and either the entity/files/directories already exist
    """


class UserStrategyError(SmartSimError):
    """Raised when there is an error with model creation inside an ensemble
    that is from a user provided permutation strategy
    """

    def __init__(self, perm_strat: str) -> None:
        message = self.create_message(perm_strat)
        super().__init__(message)

    @staticmethod
    def create_message(perm_strat: str) -> str:
        prefix = "User provided ensemble generation strategy"
        message = "failed to generate valid parameter names and values"
        return " ".join((prefix, str(perm_strat), message))


class ParameterWriterError(SmartSimError):
    """Raised in the event that input parameter files for a model
    could not be written.
    """

    def __init__(self, file_path: str, read: bool = True) -> None:
        message = self.create_message(file_path, read)
        super().__init__(message)

    @staticmethod
    def create_message(file_path: str, read: bool) -> str:
        if read:
            msg = f"Failed to read configuration file to write at {file_path}"
        else:
            msg = f"Failed to write configuration file to {file_path}"
        return msg


class SSReservedKeywordError(SmartSimError):
    """Raised when a Reserved Keyword is used incorrectly"""


class SSDBIDConflictError(SmartSimError):
    """Raised in the event that a database identifier
    is not unique when multiple databases are created
    """


class SSDBFilesNotParseable(SmartSimError):
    """Raised when the files related to the database cannot be parsed.
    Includes the case when the files do not exist.
    """


# Internal Exceptions


class SSInternalError(Exception):
    """SSInternalError is raised when an internal error is encountered"""


class SSConfigError(SSInternalError):
    """Raised when there is an error in the configuration of SmartSim"""


class LauncherError(SSInternalError):
    """Raised when there is an error in the launcher"""


class AllocationError(LauncherError):
    """Raised when there is a problem with the user WLM allocation"""


class ShellError(LauncherError):
    """Raised when error arises from function within launcher.shell
    Closely related to error from subprocess(Popen) commands
    """

    def __init__(
        self,
        message: str,
        command_list: t.Union[str, t.List[str]],
        details: t.Optional[t.Union[Exception, str]] = None,
    ) -> None:
        msg = self.create_message(message, command_list, details=details)
        super().__init__(msg)

    @staticmethod
    def create_message(
        message: str,
        command_list: t.Union[str, t.List[str]],
        details: t.Optional[t.Union[Exception, str]],
    ) -> str:
        if isinstance(command_list, list):
            command_list = " ".join(command_list)
        msg = message + "\n"
        msg += f"\nCommand: {command_list}"
        if details:
            msg += f"\nError from shell: {details}"
        return msg


class TelemetryError(SSInternalError):
    """Raised when SmartSim runs into trouble establishing or communicating
    telemetry information
    """


class UnproxyableStepError(TelemetryError):
    """Raised when a user attempts to proxy a managed ``Step`` through the
    unmanaged step proxy entry point
    """


class SmartSimCLIActionCancelled(SmartSimError):
    """Raised when a `smart` CLI command is terminated"""


class PreviewFormatError(SSUnsupportedError):
    """Raised when the output format of the preview method call is not supported"""



# BSD 2-Clause License
#
# Copyright (c) 2021-2024 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import logging
import os
import pathlib
import signal
import sys
import typing as t
from types import FrameType

import coloredlogs
import psutil

import smartsim.log
from smartsim._core.utils.helpers import decode_cmd, get_ts_ms
from smartsim._core.utils.telemetry.telemetry import write_event

STEP_PID: t.Optional[int] = None
logger = smartsim.log.get_logger(__name__)

# kill is not catchable
SIGNALS = [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGABRT]


def main(
    cmd: str,
    entity_type: str,
    cwd: str,
    status_dir: str,
) -> int:
    """This function receives an encoded step command from a SmartSim Experiment
    and runs it in a subprocess. The entrypoint integrates with the telemetry
    monitor by writing status update events. It is useful for wrapping
    unmanaged tasks - a workload manager can be queried for a managed task
    to achieve the same result.

    :param cmd: a base64 encoded cmd to execute
    :param entity_type: `SmartSimEntity` entity class. Valid values
    include: orchestrator, dbnode, ensemble, model
    :param cwd: working directory to execute the cmd from
    :param status_dir: path to the output directory for status updates
    """
    global STEP_PID  # pylint: disable=global-statement
    proxy_pid = os.getpid()

    status_path = pathlib.Path(status_dir)
    if not status_path.exists():
        status_path.mkdir(parents=True, exist_ok=True)

    if not cmd.strip():
        raise ValueError("Invalid cmd supplied")

    cleaned_cmd = decode_cmd(cmd)
    ret_code: int = 1
    logger.debug("Indirect step starting")

    start_detail = f"Proxy process {proxy_pid}"
    start_rc: t.Optional[int] = None

    try:
        process = psutil.Popen(
            cleaned_cmd,
            cwd=cwd,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        STEP_PID = process.pid
        logger.info(f"Indirect proxy {proxy_pid} child process {STEP_PID} started")
        start_detail += f" started child process {STEP_PID}"

    except Exception as ex:
        start_detail += f" failed to start child process. {ex}"
        start_rc = 1
        logger.error("Failed to create process", exc_info=True)
        cleanup()
        return 1
    finally:
        write_event(
            get_ts_ms(),
            proxy_pid,
            "",  # step_id for unmanaged task is always empty
            entity_type,
            "start",
            status_path,
            detail=start_detail,
            return_code=start_rc,
        )

    logger.info(f"Waiting for child process {STEP_PID} to complete")

    try:
        ret_code = process.wait()
    except Exception:
        logger.error("Failed to complete process", exc_info=True)
        ret_code = -1

    logger.info(
        f"Indirect proxy {proxy_pid} child process {STEP_PID} complete."
        f" return code: {ret_code}"
    )
    msg = f"Process {STEP_PID} finished with return code: {ret_code}"
    write_event(
        get_ts_ms(),
        proxy_pid,
        "",  # step_id for unmanaged task is always empty
        entity_type,
        "stop",
        status_path,
        detail=msg,
        return_code=ret_code,
    )
    cleanup()

    return ret_code


def cleanup() -> None:
    """Perform cleanup required for clean termination"""
    global STEP_PID  # pylint: disable=global-statement
    if STEP_PID is None:
        return

    logger.info("Performing cleanup")

    try:
        # attempt to stop the subprocess performing step-execution
        if psutil.pid_exists(STEP_PID):
            process = psutil.Process(STEP_PID)
            process.terminate()
    except psutil.NoSuchProcess:
        # swallow exception to avoid overwriting outputs from cmd
        ...

    except OSError as ex:
        logger.warning(f"Failed to clean up step executor gracefully: {ex}")
    finally:
        STEP_PID = None


def handle_signal(signo: int, _frame: t.Optional[FrameType]) -> None:
    """Helper function to ensure clean process termination"""
    logger.info(f"handling signal {signo}")
    if not signo:
        logger.warning("Received signal with no signo")

    cleanup()


def register_signal_handlers() -> None:
    """Register a signal handling function for all termination events"""
    for sig in SIGNALS:
        signal.signal(sig, handle_signal)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prefix_chars="+", description="SmartSim Step Executor"
    )
    parser.add_argument(
        "+name", type=str, help="Name of the step being executed", required=True
    )
    parser.add_argument(
        "+command", type=str, help="The command to execute", required=True
    )
    parser.add_argument(
        "+entity_type",
        type=str,
        help="The type of entity related to the step",
        required=True,
    )
    parser.add_argument(
        "+working_dir",
        type=str,
        help="The working directory of the executable",
        required=True,
    )
    parser.add_argument(
        "+telemetry_dir",
        type=str,
        help="Directory for telemetry output",
        required=True,
    )
    return parser


if __name__ == "__main__":
    arg_parser = get_parser()
    os.environ["PYTHONUNBUFFERED"] = "1"
    parsed_args = arg_parser.parse_args()

    # Set up a local private logger for when this module is run as an entry point
    level = logger.getEffectiveLevel()
    logger = logging.getLogger(f"{__name__}.{parsed_args.name}")
    logger.propagate = False
    logger.setLevel(level)

    fh = logging.FileHandler(f"{parsed_args.name}.indirect.log")
    coloredlogs.HostNameFilter.install(fh)
    fh.setFormatter(
        logging.Formatter(
            smartsim.log.DEFAULT_LOG_FORMAT,
            datefmt=smartsim.log.DEFAULT_DATE_FORMAT,
        )
    )
    logger.addHandler(fh)

    try:
        logger.debug("Starting indirect step execution")

        # make sure to register the cleanup before the start the process
        # so our signaller will be able to stop the database process.
        register_signal_handlers()

        rc = main(
            cmd=parsed_args.command,
            entity_type=parsed_args.entity_type,
            cwd=parsed_args.working_dir,
            status_dir=parsed_args.telemetry_dir,
        )
        sys.exit(rc)

    # gracefully exit the processes in the distributed application that
    # we do not want to have start a colocated process. Only one process
    # per node should be running.
    except Exception as e:
        logger.exception(f"An unexpected error caused step execution to fail: {e}")
        sys.exit(1)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024 Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import json
import os
import signal
import textwrap
import typing as t
from subprocess import PIPE, STDOUT
from types import FrameType

import psutil

from smartsim._core.utils.network import current_ip
from smartsim.entity.dbnode import LaunchedShardData
from smartsim.log import get_logger

logger = get_logger(__name__)

"""
Redis/KeyDB entrypoint script
"""

DBPID: t.Optional[int] = None

# kill is not catchable
SIGNALS = [signal.SIGINT, signal.SIGQUIT, signal.SIGTERM, signal.SIGABRT]


def handle_signal(signo: int, _frame: t.Optional[FrameType]) -> None:
    if not signo:
        logger.warning("Received signal with no signo")
    cleanup()


def build_bind_args(source_addr: str, *addrs: str) -> t.Tuple[str, ...]:
    return (
        "--bind",
        source_addr,
        *addrs,
        # pin source address to avoid random selection by Redis
        "--bind-source-addr",
        source_addr,
    )


def build_cluster_args(shard_data: LaunchedShardData) -> t.Tuple[str, ...]:
    if cluster_conf_file := shard_data.cluster_conf_file:
        return ("--cluster-enabled", "yes", "--cluster-config-file", cluster_conf_file)
    return ()


def print_summary(
    cmd: t.List[str], network_interface: str, shard_data: LaunchedShardData
) -> None:
    print(
        textwrap.dedent(f"""\
            ----------- Running Command ----------
            COMMAND: {' '.join(cmd)}
            IPADDRESS: {shard_data.hostname}
            NETWORK: {network_interface}
            SMARTSIM_ORC_SHARD_INFO: {json.dumps(shard_data.to_dict())}
            --------------------------------------

            --------------- Output ---------------

            """),
        flush=True,
    )


def main(args: argparse.Namespace) -> int:
    global DBPID  # pylint: disable=global-statement

    src_addr, *bind_addrs = (current_ip(net_if) for net_if in args.ifname.split(","))
    shard_data = LaunchedShardData(
        name=args.name, hostname=src_addr, port=args.port, cluster=args.cluster
    )

    cmd = [
        args.orc_exe,
        args.conf_file,
        *args.rai_module,
        "--port",
        str(args.port),
        *build_cluster_args(shard_data),
        *build_bind_args(src_addr, *bind_addrs),
    ]

    print_summary(cmd, args.ifname, shard_data)

    try:
        process = psutil.Popen(cmd, stdout=PIPE, stderr=STDOUT)
        DBPID = process.pid

        for line in iter(process.stdout.readline, b""):
            print(line.decode("utf-8").rstrip(), flush=True)
    except Exception:
        cleanup()
        logger.error("Database process starter raised an exception", exc_info=True)
        return 1
    return 0


def cleanup() -> None:
    logger.debug("Cleaning up database instance")
    try:
        # attempt to stop the database process
        if DBPID is not None:
            psutil.Process(DBPID).terminate()
    except psutil.NoSuchProcess:
        logger.warning("Couldn't find database process to kill.")
    except OSError as e:
        logger.warning(f"Failed to clean up database gracefully: {str(e)}")


if __name__ == "__main__":
    os.environ["PYTHONUNBUFFERED"] = "1"

    parser = argparse.ArgumentParser(
        prefix_chars="+", description="SmartSim Process Launcher"
    )
    parser.add_argument(
        "+orc-exe", type=str, help="Path to the orchestrator executable", required=True
    )
    parser.add_argument(
        "+conf-file",
        type=str,
        help="Path to the orchestrator configuration file",
        required=True,
    )
    parser.add_argument(
        "+rai-module",
        nargs="+",
        type=str,
        help=(
            "Command for the orcestrator to load the Redis AI module with "
            "symbols seperated by whitespace"
        ),
        required=True,
    )
    parser.add_argument(
        "+name", type=str, help="Name to identify the shard", required=True
    )
    parser.add_argument(
        "+port",
        type=int,
        help="The port on which to launch the shard of the orchestrator",
        required=True,
    )
    parser.add_argument(
        "+ifname", type=str, help="Network Interface name", required=True
    )
    parser.add_argument(
        "+cluster",
        action="store_true",
        help="Specify if this orchestrator shard is part of a cluster",
    )

    args_ = parser.parse_args()

    # make sure to register the cleanup before the start
    # the process so our signaller will be able to stop
    # the database process.
    for sig in SIGNALS:
        signal.signal(sig, handle_signal)

    raise SystemExit(main(args_))



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import time
import typing as t
from shutil import which

from ....error import LauncherError
from ....log import get_logger
from ....settings import (
    MpiexecSettings,
    MpirunSettings,
    OrterunSettings,
    RunSettings,
    SbatchSettings,
    SettingsBase,
    SrunSettings,
)
from ....status import SmartSimStatus
from ...config import CONFIG
from ..launcher import WLMLauncher
from ..step import (
    LocalStep,
    MpiexecStep,
    MpirunStep,
    OrterunStep,
    SbatchStep,
    SrunStep,
    Step,
)
from ..stepInfo import SlurmStepInfo, StepInfo
from .slurmCommands import sacct, scancel, sstat
from .slurmParser import parse_sacct, parse_sstat_nodes, parse_step_id_from_sacct

logger = get_logger(__name__)


class SlurmLauncher(WLMLauncher):
    """This class encapsulates the functionality needed
    to launch jobs on systems that use Slurm as a workload manager.

    All WLM launchers are capable of launching managed and unmanaged
    jobs. Managed jobs are queried through interaction with with WLM,
    in this case Slurm. Unmanaged jobs are held in the TaskManager
    and are managed through references to their launching process ID
    i.e. a psutil.Popen object
    """

    # init in launcher.py (WLMLauncher)

    # RunSettings types supported by this launcher
    @property
    def supported_rs(self) -> t.Dict[t.Type[SettingsBase], t.Type[Step]]:
        # RunSettings types supported by this launcher
        return {
            SrunSettings: SrunStep,
            SbatchSettings: SbatchStep,
            MpirunSettings: MpirunStep,
            MpiexecSettings: MpiexecStep,
            OrterunSettings: OrterunStep,
            RunSettings: LocalStep,
        }

    def get_step_nodes(self, step_names: t.List[str]) -> t.List[t.List[str]]:
        """Return the compute nodes of a specific job or allocation

        This function returns the compute nodes of a specific job or allocation
        in a list with the duplicates removed.

        Output gleaned from sstat e.g. the following

        29917893.extern|nid00034|44860|
        29917893.0|nid00034|44887,45151,45152,45153,45154,45155|
        29917893.2|nid00034|45174|

        would return nid00034

        :param step_names: list of job step names
        :raises LauncherError: if nodelist aquisition fails
        :return: list of hostnames
        """
        _, step_ids = self.step_mapping.get_ids(step_names, managed=True)
        step_str = _create_step_id_str([val for val in step_ids if val is not None])
        output, _ = sstat([step_str, "-i", "-n", "-p", "-a"], raise_on_err=True)

        # parse node list for each step
        node_lists = []
        for step_id in step_ids:
            node_lists.append(parse_sstat_nodes(output, step_id or ""))

        if len(node_lists) < 1:
            raise LauncherError("Failed to retrieve nodelist from stat")
        return node_lists

    def run(self, step: Step) -> t.Optional[str]:
        """Run a job step through Slurm

        :param step: a job step instance
        :raises LauncherError: if launch fails
        :return: job step id if job is managed
        """
        self.check_for_slurm()
        if not self.task_manager.actively_monitoring:
            self.task_manager.start()

        cmd_list = step.get_launch_cmd()
        step_id = None
        task_id = None

        # Launch a batch step with Slurm
        if isinstance(step, SbatchStep):
            # wait for batch step to submit successfully
            return_code, out, err = self.task_manager.start_and_wait(cmd_list, step.cwd)
            if return_code != 0:
                raise LauncherError(f"Sbatch submission failed\n {out}\n {err}")
            if out:
                step_id = out.strip()
                logger.debug(f"Gleaned batch job id: {step_id} for {step.name}")

        # Launch a in-allocation or on-allocation (if srun) command
        elif isinstance(step, SrunStep):
            task_id = self.task_manager.start_task(cmd_list, step.cwd)
        else:
            # MPI/local steps don't direct output like slurm steps
            out, err = step.get_output_files()

            # pylint: disable-next=consider-using-with
            output = open(out, "w+", encoding="utf-8")
            # pylint: disable-next=consider-using-with
            error = open(err, "w+", encoding="utf-8")
            task_id = self.task_manager.start_task(
                cmd_list, step.cwd, step.env, out=output.fileno(), err=error.fileno()
            )

        if not step_id and step.managed:
            step_id = self._get_slurm_step_id(step)

        self.step_mapping.add(step.name, step_id, task_id, step.managed)

        # give slurm a rest
        # TODO make this configurable
        time.sleep(1)

        return step_id

    def stop(self, step_name: str) -> StepInfo:
        """Step a job step

        :param step_name: name of the job to stop
        :return: update for job due to cancel
        """
        stepmap = self.step_mapping[step_name]
        if stepmap.managed:
            step_id = str(stepmap.step_id)
            # Check if step_id is part of colon-separated run,
            # this is reflected in a '+' in the step id,
            # so that the format becomes 12345+1.0.
            # If we find it it can mean two things:
            # a MPMD srun command, or a heterogeneous job.
            # If it is a MPMD srun, then stop parent step because
            # sub-steps cannot be stopped singularly.
            sub_step = "+" in step_id
            het_job = os.getenv("SLURM_HET_SIZE") is not None
            # If it is a het job, we can stop
            # them like this. Slurm will throw an error, but
            # will actually kill steps correctly.
            if sub_step and not het_job:
                step_id = step_id.split("+", maxsplit=1)[0]
            scancel_rc, _, err = scancel([step_id])
            if scancel_rc != 0:
                if het_job:
                    msg = (
                        "SmartSim received a non-zero exit code while canceling"
                        f" a heterogeneous job step {step_name}!\n"
                        "The following error might be internal to Slurm\n"
                        "and the heterogeneous job step could have been correctly"
                        " canceled.\nSmartSim will consider it canceled.\n"
                    )
                else:
                    msg = f"Unable to cancel job step {step_name}\n{err}"
                logger.warning(msg)
            if stepmap.task_id:
                self.task_manager.remove_task(str(stepmap.task_id))
        else:
            self.task_manager.remove_task(str(stepmap.task_id))

        _, step_info = self.get_step_update([step_name])[0]
        if not step_info:
            raise LauncherError(f"Could not get step_info for job step {step_name}")

        step_info.status = (
            SmartSimStatus.STATUS_CANCELLED
        )  # set status to cancelled instead of failed
        return step_info

    @staticmethod
    def _get_slurm_step_id(step: Step, interval: int = 2) -> str:
        """Get the step_id of a step from sacct

        Parses sacct output by looking for the step name
        e.g. the following

        SmartSim|119225|
        extern|119225.extern|
        m1-119225.0|119225.0|
        m2-119225.1|119225.1|
        """
        time.sleep(interval)
        step_id: t.Optional[str] = None
        trials = CONFIG.wlm_trials
        while trials > 0:
            output, _ = sacct(
                ["--noheader", "-p", "--format=jobname,jobid"], raise_on_err=True
            )

            step_id = parse_step_id_from_sacct(output, step.name)
            if step_id:
                break
            else:
                time.sleep(interval)
                trials -= 1
        if not step_id:
            raise LauncherError("Could not find id of launched job step")
        return step_id

    def _get_managed_step_update(self, step_ids: t.List[str]) -> t.List[StepInfo]:
        """Get step updates for WLM managed jobs

        :param step_ids: list of job step ids
        :return: list of updates for managed jobs
        """
        step_str = _create_step_id_str(step_ids)
        sacct_out, _ = sacct(
            ["--noheader", "-p", "-b", "--jobs", step_str], raise_on_err=True
        )

        # (status, returncode)
        stat_tuples = [parse_sacct(sacct_out, step_id) for step_id in step_ids]

        # create SlurmStepInfo objects to return
        updates: t.List[StepInfo] = []
        for stat_tuple, step_id in zip(stat_tuples, step_ids):
            _rc = int(stat_tuple[1]) if stat_tuple[1] else None
            info = SlurmStepInfo(stat_tuple[0], _rc)

            task_id = self.step_mapping.get_task_id(step_id)
            if task_id:
                # we still check the task manager for jobs that didn't ever
                # become a fully managed job (e.g. error in slurm arguments)
                tid = str(task_id)
                _, return_code, out, err = self.task_manager.get_task_update(tid)
                if return_code and return_code != 0:
                    # tack on Popen error and output to status update.
                    info.output = out
                    info.error = err

            updates.append(info)
        return updates

    @staticmethod
    def check_for_slurm() -> None:
        """Check if slurm is available

        This function checks for slurm commands where the experiment
        is bring run

        :raises LauncherError: if no access to slurm
        """
        if not which("sbatch") and not which("sacct"):
            error = (
                "User attempted Slurm methods without access to Slurm "
                "at the call site.\n"
            )
            raise LauncherError(error)

    def __str__(self) -> str:
        return "Slurm"


def _create_step_id_str(step_ids: t.List[str]) -> str:
    return ",".join(step_ids)



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools
import logging
import pathlib
import socket
import sys
import threading
import typing as t
from contextvars import ContextVar, copy_context

import coloredlogs

from smartsim._core.config import CONFIG

# constants
DEFAULT_DATE_FORMAT: t.Final[str] = "%H:%M:%S"
DEFAULT_LOG_FORMAT: t.Final[str] = (
    "%(asctime)s %(hostname)s %(name)s[%(process)d:%(threadName)s] "
    "%(levelname)s %(message)s"
)
EXPERIMENT_LOG_FORMAT = DEFAULT_LOG_FORMAT.replace("s[%", "s {%(exp_path)s} [%")

# configure colored loggs
coloredlogs.DEFAULT_DATE_FORMAT = DEFAULT_DATE_FORMAT
coloredlogs.DEFAULT_LOG_FORMAT = DEFAULT_LOG_FORMAT

# create context vars used by loggers
ctx_exp_path = ContextVar("exp_path", default="")


# Generic types for method contextualizers
_T = t.TypeVar("_T")
_RT = t.TypeVar("_RT")
_ContextT = t.TypeVar("_ContextT")

if t.TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec

    _PR = ParamSpec("_PR")


def _translate_log_level(user_log_level: str = "info") -> str:
    """Translate value of CONFIG.log_level to one
    accepted as ``level`` option by Python's logging module.

       Logging levels
         - quiet: Just shows errors and warnings
         - info: Show basic information and errors (default)
         - debug: Shows info, errors and user debug information
         - developer: Shows everything happening during execution
                      extremely verbose logging.

    :param user_log_level: log level specified by user, defaults to info
    :returns: Log level for coloredlogs
    """
    user_log_level = user_log_level.lower()
    if user_log_level in ["info", "debug", "warning"]:
        return user_log_level
    if user_log_level == "quiet":
        return "warning"
    # extremely verbose logging used internally
    if user_log_level == "developer":
        return "debug"
    return "info"


def get_exp_log_paths() -> t.Tuple[t.Optional[pathlib.Path], t.Optional[pathlib.Path]]:
    """Returns the output and error file paths to experiment logs.
    Returns None for both paths if experiment context is unavailable.

    :returns: 2-tuple of paths to experiment logs in form (output_path, error_path)
    """
    default_paths = None, None

    if _path := ctx_exp_path.get():
        file_out = pathlib.Path(_path) / CONFIG.telemetry_subdir / "logs/smartsim.out"
        file_err = pathlib.Path(_path) / CONFIG.telemetry_subdir / "logs/smartsim.err"
        return file_out, file_err

    return default_paths


class ContextThread(threading.Thread):
    """Thread that ensures the context vars of the caller are available"""

    def run(self) -> None:
        """Execute a thread on a copy of the current thread context"""
        ctx = copy_context()
        return ctx.run(super().run)


class ContextInjectingLogFilter(logging.Filter):
    """Filter that performs enrichment of a log record by adding context
    information about the experiment being executed"""

    def filter(self, record: logging.LogRecord) -> bool:
        """Enrich log records with active experiment context

        :param record: the record to evaluate for filtering
        :returns: always True
        """
        record.exp_path = ctx_exp_path.get()
        return True


class HostnameFilter(logging.Filter):
    """Filter that performs enrichment of a log record by adding
    the hostname of the machine executing the code"""

    def __init__(self, name: str = "") -> None:
        super().__init__(name)
        self._hostname = ""

    @property
    @functools.lru_cache
    def hostname(self) -> str:
        """Returns the hostname of the machine executing the code"""
        self._hostname = socket.gethostname()
        return self._hostname

    def filter(self, record: logging.LogRecord) -> bool:
        # the hostname may already added if using the `ColoredLogs` plugin
        if not hasattr(record, "hostname"):
            record.hostname = self.hostname
        return True


class ContextAwareLogger(logging.Logger):
    """A logger customized to automatically write experiment logs to a
    dynamic target directory by inspecting the value of a context var"""

    def __init__(self, name: str, level: t.Union[int, str] = 0) -> None:
        super().__init__(name, level)
        self.addFilter(ContextInjectingLogFilter(name="exp-ctx-log-filter"))

    def _log(
        self,
        level: int,
        msg: object,
        args: t.Any,
        exc_info: t.Optional[t.Any] = None,
        extra: t.Optional[t.Any] = None,
        stack_info: bool = False,
        stacklevel: int = 1,
    ) -> None:
        """Automatically attach file handlers if contextual information is found"""
        file_out, file_err = get_exp_log_paths()

        if not all([file_out, file_err]):
            super()._log(level, msg, args, exc_info, extra, stack_info, stacklevel)
            return

        _lvl = logging.getLevelName(self.level)
        fmt = EXPERIMENT_LOG_FORMAT

        low_pass = LowPassFilter(_lvl)
        h_out = log_to_exp_file(str(file_out), self, _lvl, fmt, low_pass)
        h_err = log_to_exp_file(str(file_err), self, "WARN", fmt)

        super()._log(level, msg, args, exc_info, extra, stack_info, stacklevel)

        for handler in [h_out, h_err]:
            self.removeHandler(handler)


def get_logger(
    name: str, log_level: t.Optional[str] = None, fmt: t.Optional[str] = None
) -> logging.Logger:
    """Return a logger instance

    levels:
        - quiet
        - info
        - debug
        - developer

    examples:
        # returns a logger with the name of the module
        logger = get_logger(__name__)

        logger.info("This is a message")
        logger.debug("This is a debug message")
        logger.error("This is an error message")
        logger.warning("This is a warning message")

    :param name: the name of the desired logger
    :param log_level: what level to set the logger to
    :param fmt: the format of the log messages
    :returns: logger instance
    """
    # if name is None, then logger is the root logger
    # if not root logger, get the name of file without prefix.
    user_log_level = CONFIG.log_level
    if user_log_level != "developer":
        name = "SmartSim"

    logging.setLoggerClass(ContextAwareLogger)
    logger = logging.getLogger(name)
    if log_level:
        logger.setLevel(log_level)
    else:
        log_level = _translate_log_level(user_log_level)
    coloredlogs.install(level=log_level, logger=logger, fmt=fmt, stream=sys.stdout)
    return logger


class LowPassFilter(logging.Filter):
    """A filter that passes all records below a specified level"""

    def __init__(self, maximum_level: str = "INFO"):
        """Create a low-pass log filter allowing messages below a specific log level

        :param maximum_level: The maximum log level to be passed by the filter
        """
        super().__init__()
        self.max = maximum_level

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter log records; pass those less than or equal to the maximum level

        :param record: the record to evaluate for filtering
        :returns: True if record level passes filter, False otherwise
        """
        # If a string representation of the level is passed in,
        # the corresponding numeric value is returned.
        level_no: int = logging.getLevelName(self.max)
        return record.levelno <= level_no


def log_to_file(filename: str, log_level: str = "debug") -> None:
    """Installs a second filestream handler to the root logger,
    allowing subsequent logging calls to be sent to filename.

    :param filename: the name of the desired log file.
    :param log_level: as defined in get_logger.  Can be specified
                      to allow the file to store more or less verbose
                      logging information.
    """
    logger = logging.getLogger("SmartSim")
    stream = open(  # pylint: disable=consider-using-with
        filename, "w+", encoding="utf-8"
    )
    coloredlogs.install(stream=stream, logger=logger, level=log_level)


def log_to_exp_file(
    filename: str,
    logger: logging.Logger,
    log_level: str = "warn",
    fmt: t.Optional[str] = EXPERIMENT_LOG_FORMAT,
    log_filter: t.Optional[logging.Filter] = None,
) -> logging.Handler:
    """Installs a second filestream handler to the root logger,
    allowing subsequent logging calls to be sent to filename.

    :param filename: the name of the desired log file.
    :param log_level: as defined in get_logger.  Can be specified
                      to allow the file to store more or less verbose
                      logging information.
    :param logger: an existing logger to add the handler to
    :param fmt: a log format for the handler (otherwise, EXPERIMENT_LOG_FORMAT)
    :param log_filter: log filter to attach to handler
    :return: logging.Handler
    """
    # ensure logs are written even if specified dir doesn't exist
    log_path = pathlib.Path(filename)
    if not log_path.parent.exists():
        log_path.parent.mkdir(parents=True, exist_ok=True)

    handler = logging.FileHandler(filename, mode="a+", encoding="utf-8")

    if log_filter:
        handler.addFilter(log_filter)

    formatter = logging.Formatter(fmt=fmt, datefmt=DEFAULT_DATE_FORMAT)

    handler.setFormatter(formatter)
    handler.setLevel(log_level.upper())

    logger.addHandler(handler)
    return handler


def method_contextualizer(
    ctx_var: ContextVar[_ContextT],
    ctx_map: t.Callable[[_T], _ContextT],
) -> """t.Callable[
    [t.Callable[Concatenate[_T, _PR], _RT]],
    t.Callable[Concatenate[_T, _PR], _RT],
]""":
    """Parameterized-decorator factory that enables a target value
    to be placed into global context prior to execution of the
    decorated method.
    Usage Note: the use of `self` below requires that the decorated function is passed
    the object containing a value that will be modified in the context. `ctx_map`
    must accept an instance of matching type.

    :param ctx_var: The ContextVar that will be modified
    :param ctx_map: A function that returns the value to be set to ctx_var
    """

    def _contextualize(
        fn: "t.Callable[Concatenate[_T, _PR], _RT]", /
    ) -> "t.Callable[Concatenate[_T, _PR], _RT]":
        """Executes the decorated method in a cloned context and ensures
        `ctx_var` is updated to the value returned by `ctx_map` prior to
        calling the decorated method"""

        @functools.wraps(fn)
        def _contextual(
            self: _T,
            *args: "_PR.args",
            **kwargs: "_PR.kwargs",
        ) -> _RT:
            """A decorator operator that runs the decorated method in a new
            context with the desired contextual information modified."""

            def _ctx_modifier() -> _RT:
                """Helper to simplify calling the target method with the
                modified value set in `ctx_var`"""
                ctx_val = ctx_map(self)
                token = ctx_var.set(ctx_val)
                result = fn(self, *args, **kwargs)
                ctx_var.reset(token)
                return result

            ctx = copy_context()
            return ctx.run(_ctx_modifier)

        return _contextual

    return _contextualize



# BSD 2-Clause License
#
# Copyright (c) 2021-2024, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import typing as t

from pydantic import BaseModel, Field

import smartsim._core.schemas.utils as _utils
from smartsim.status import SmartSimStatus

# Black and Pylint disagree about where to put the `...`
# pylint: disable=multiple-statements

response_registry = _utils.SchemaRegistry["DragonResponse"]()


class DragonResponse(BaseModel):
    error_message: t.Optional[str] = None


@response_registry.register("run")
class DragonRunResponse(DragonResponse):
    step_id: t.Annotated[str, Field(min_length=1)]


@response_registry.register("status_update")
class DragonUpdateStatusResponse(DragonResponse):
    # status is a dict: {step_id: (is_alive, returncode)}
    statuses: t.Mapping[
        t.Annotated[str, Field(min_length=1)],
        t.Tuple[SmartSimStatus, t.Optional[t.List[int]]],
    ] = {}


@response_registry.register("stop")
class DragonStopResponse(DragonResponse): ...


@response_registry.register("handshake")
class DragonHandshakeResponse(DragonResponse):
    dragon_pid: int


@response_registry.register("bootstrap")
class DragonBootstrapResponse(DragonResponse):
    dragon_pid: int


@response_registry.register("shutdown")
class DragonShutdownResponse(DragonResponse): ...
